/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "io_spec.h"
#include "buffer.h"
#include "linklist.h"
#include "util.h"



struct BufferSet {
  int NumBufs; /* size of array */
  int NextToFill;
  int NextToEmpty;
  int fd; /* corresponding file descriptor */
  struct SBuf **Bufs; 
};

struct BufferSet *Bufs; /* Pointer to array of buffers, 1 per input */
int NumInputs=2;

int NumBufs=12;

int NumChannels;
int SampleRate;
int BlockSize;
int Subtract=0;


int GotMagic=0;


int readpipe=0;
int writepipe=1;

static char **global_argv;

int OneBlockAvailable(void)
     /* Returns one when there is at least one full block available 
	from each input */
     
{
  int Cnt;

  for (Cnt=0;Cnt < NumInputs; Cnt++) {
    if (Bufs[Cnt].NextToFill==Bufs[Cnt].NextToEmpty)
      return 0; /* not a complete buffer */
  }
  return 1;
}


int main(int argc,char *argv[])
{
  int Cnt,Cnt2;
  struct SBuf *OutBuf;
  struct StatusBlk *Status;
  fd_set readfd,writefd;
  int maxfd;


  global_argv=argv;

  for (Cnt=1;argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    if (!strncmp(argv[Cnt],"-numinputs=",11)) {
      NumInputs=strtol(argv[Cnt]+11,NULL,10);
    }
    if (!strcmp(argv[Cnt],"-subtract")) {
      Subtract=1;
    }
    if (!strncmp(argv[Cnt],"-numbuffers=",12)) {
      NumBufs=strtol(argv[Cnt]+12,NULL,10);
    }
  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   input 0 \n"
	    "        #1   aglaophone format output\n"
	    "        #2   error messages (do not connect)\n"
	    "        #3..(n+1)  inputs 1..(n-1), where n=numinputs\n"
      	    "\n"
	    "Usage:\n"
	    "   %s -magic -numinputs=2 -numbuffers=12 -subtract\n"
	    "\n"
	    "   Default is 2 inputs, 12 buffers\n"
	    "   subtract negates all inputs but #0\n"
	    ,argv[0]);
    exit(1);
  }
  
  Bufs=malloc(sizeof(struct BufferSet)*NumInputs);
  
  for (Cnt=0;Cnt < NumInputs; Cnt++) {
    Bufs[Cnt].NumBufs=NumBufs;
    Bufs[Cnt].NextToFill=0;
    Bufs[Cnt].NextToEmpty=0;
    if (!Cnt)
      Bufs[Cnt].fd=0;
    else Bufs[Cnt].fd=Cnt+2;
    
    Bufs[Cnt].Bufs=malloc(sizeof(struct SBuf *)*Bufs[Cnt].NumBufs);
    for (Cnt2=0;Cnt2 < Bufs[Cnt].NumBufs;Cnt2++) {
      Bufs[Cnt].Bufs[Cnt2]=CreateSBuf();
    }
    SBufPrepareHeader(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill]);
    setnonblocking(Bufs[Cnt].fd);
  }

  OutBuf=CreateSBuf();
  
  for (;;) {
    
    /* Read all inputs */

    do {
      maxfd=2;
      FD_ZERO(&readfd);
    
      /* determine which inputs on which to accept input */
      for (Cnt=0;Cnt < NumInputs;Cnt++) {
	if (((Bufs[Cnt].NextToEmpty != ((Bufs[Cnt].NextToFill+1)%Bufs[Cnt].NumBufs))) && SBufIsReady(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill])) {
	  FD_SET(Bufs[Cnt].fd,&readfd);
	  if (Bufs[Cnt].fd > maxfd) maxfd=Bufs[Cnt].fd;
	}
      }
      
      select(maxfd+1,&readfd,NULL,NULL,NULL);
      /* read in data */
      for (Cnt=0;Cnt < NumInputs;Cnt++) {
	if (((Bufs[Cnt].NextToEmpty != ((Bufs[Cnt].NextToFill+1)%Bufs[Cnt].NumBufs))) && SBufIsReady(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill])) {
	  ReadSBuf(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill],Bufs[Cnt].fd);
	  /*	  fprintf(stderr,"reading socket #%d\n",(int)Bufs[Cnt].fd);*/
	  if (SBufIsHeaderDone(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill])) {
	    assert(SBufCheckHeader(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill]));
	    SBufReallocateIfNeeded(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill]);
	    SBufPrepareFinish(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill]);
	    
	  }
	  if (SBufIsDataDone(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill])) {
	    /* Block finished */
	    /* Increment next-block-to-fill */
	    Bufs[Cnt].NextToFill=(Bufs[Cnt].NextToFill+1)%Bufs[Cnt].NumBufs;
	    
	    
	    /* Get next block ready */
	    SBufPrepareHeader(Bufs[Cnt].Bufs[Bufs[Cnt].NextToFill]);
	  }
	}
      }
    } while (!OneBlockAvailable());
    
    /* make sure all the blocksizes, etc. match */
    SampleRate=Bufs[0].Bufs[Bufs[0].NextToEmpty]->Header->SampleRate;
    NumChannels=Bufs[0].Bufs[Bufs[0].NextToEmpty]->Header->NumChannels;
    BlockSize=Bufs[0].Bufs[Bufs[0].NextToEmpty]->Header->BlockSize;
    for (Cnt=1;Cnt < NumInputs;Cnt++) {
      if (Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Header->SampleRate != SampleRate) {
	fprintf(stderr,"%s: Sample rate mismatch, %d and %d\n",argv[0],(int)Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Header->SampleRate,(int)SampleRate);
	exit(1);
      }
      if (Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Header->NumChannels != NumChannels) {
	fprintf(stderr,"%s: Number of channels mismatch, %d and %d\n",argv[0],(int)Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Header->NumChannels,(int)NumChannels);
	exit(1);
	
      }
      if (Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Header->BlockSize != BlockSize) {
	fprintf(stderr,"%s: Block size mismatch, %d and %d\n",argv[0],(int)Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Header->BlockSize,(int)BlockSize);
	exit(1);
	
      }
      
    }
    
    
    /* OK. We've got all the data for one block. Let's 
       perform the addition operation */

    SBufSetHeader(OutBuf,NumChannels,SampleRate,Bufs[0].Bufs[Bufs[0].NextToEmpty]->Header->FrameNumber,BlockSize);
    SBufReallocateIfNeeded(OutBuf);
    
    /* Clear buffer */
    for (Cnt=0;Cnt < BlockSize*NumChannels;Cnt++) {
      OutBuf->Data[Cnt]=0;
    }

    if (!Subtract) {
      /* add */
      for (Cnt=0;Cnt < NumInputs;Cnt++) {
	for (Cnt2=0;Cnt2 < BlockSize *NumChannels;Cnt2++) {
	  OutBuf->Data[Cnt2]+=Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Data[Cnt2];
	}
      }
    } else {
      /* subtract */
      for (Cnt=0;Cnt < 1;Cnt++) {
	for (Cnt2=0;Cnt2 < BlockSize *NumChannels;Cnt2++) {
	  OutBuf->Data[Cnt2]+=Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Data[Cnt2];
	}
      }
      for (;Cnt < NumInputs;Cnt++) {
	for (Cnt2=0;Cnt2 < BlockSize *NumChannels;Cnt2++) {
	  OutBuf->Data[Cnt2]-=Bufs[Cnt].Bufs[Bufs[Cnt].NextToEmpty]->Data[Cnt2];
	}
      }
      

    }
    /* now write it out (this is a blocking operation) */
    SBufPrepare(OutBuf);
    WriteSBuf(OutBuf,writepipe);

    /* Increment NextToEmpty */
    for (Cnt=0;Cnt < NumInputs;Cnt++) {
      Bufs[Cnt].NextToEmpty=(Bufs[Cnt].NextToEmpty+1)%Bufs[Cnt].NumBufs;
    }
    
  }

  
}

