/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This module provides upsampling followed by filtering */
/* Filtering is done in the time domain with integer operations */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "io_spec.h"
#include "buffer.h"
#include "linklist.h"
#include "util.h"
#include "filterdownsample.h"


int GotMagic=0;

int BlockSize=0;
int NumChannels=0;
int SampleRate=0;
int NoPanel=0;
char *WindowTitle="Filter and downsample";

int readpipe=0;
int writepipe=1;

int downsamplefactor=2;

char *filterfile=NULL;

static char **global_argv;

float *float_filter=NULL;
INT32 *int_filter=NULL;
int filterlength;

/* largely ripped off from soundcardio/filter.c */

struct FilterControlBlock {
  /* all data is private */
  int SampleFactor;
  int NumChannels;
  INT32 *WorkBuf; /* Working buffer for sample processing */
  int WorkBufSize; /* number of accumulated samples */
};

void ClearFCB(struct FilterControlBlock *FCB) /* zero out any stored response in the FCB */
{
  if (FCB->WorkBuf && FCB->WorkBufSize)
    memset(FCB->WorkBuf,0,FCB->WorkBufSize*FCB->NumChannels*sizeof(INT32));
  
}


struct FilterControlBlock *FilterAndDownsample(INT16 *InputBuf,INT16 *OutputBuf,int DownsampleAmnt,int NumChannels,int InputBufferSize,struct FilterControlBlock *FCB,INT32 *filter,int filterlength)
{
  int OutputBufferSize;
  int Size;
  int Cnt,Cnt2,Channel;
  INT16 Value;
  INT32 *Ptr;


  assert((InputBufferSize % DownsampleAmnt)==0);
  OutputBufferSize=InputBufferSize/DownsampleAmnt;


  if (!FCB) {
    FCB=malloc(sizeof(struct FilterControlBlock));
    FCB->SampleFactor=DownsampleAmnt;
    FCB->NumChannels=NumChannels;
    FCB->WorkBuf=NULL;
    FCB->WorkBufSize=0;
  }
  
  
  
  if (FCB->SampleFactor != DownsampleAmnt || FCB->NumChannels != NumChannels) FCB->WorkBufSize=0;
  
  FCB->SampleFactor=DownsampleAmnt;
  FCB->NumChannels=NumChannels;
    
  
  Size=OutputBufferSize+filterlength/DownsampleAmnt+1;

  if (FCB->WorkBufSize != Size) {
    if (FCB->WorkBuf) free(FCB->WorkBuf);
    FCB->WorkBufSize=Size;
    /* allocate and fill with zeros */
    FCB->WorkBuf=calloc(FCB->WorkBufSize*FCB->NumChannels*sizeof(INT32),1);
  }

  Ptr=FCB->WorkBuf;

  /* Do the filtering */

  for (Cnt=0;Cnt < InputBufferSize;Cnt++) {
    for (Channel=0;Channel < NumChannels; Channel++) {
      Value=InputBuf[Cnt*NumChannels+Channel];
      for (Cnt2=((Cnt%DownsampleAmnt)==0) ? 0:(DownsampleAmnt-(Cnt%DownsampleAmnt));Cnt2 < filterlength;Cnt2+=DownsampleAmnt) {
	Ptr[(Cnt+Cnt2)/DownsampleAmnt*NumChannels+Channel]+=Value*filter[Cnt2];
      }
    }
  }
  
  /* done the filtering. Now extract the filtered data into the outputbuf, and move the remains
     into the start of the WorkBuf, where it will be used by the next filtering operation */
  
  for (Cnt=0; Cnt < OutputBufferSize;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      OutputBuf[Cnt*NumChannels+Channel]=Ptr[Cnt*NumChannels+Channel] >> 15; /* only take high-order 17 bits */
    }
  }
  
  
  /* Move tail end: */
  
  for (Cnt=OutputBufferSize;Cnt < Size;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      Ptr[(Cnt-OutputBufferSize)*NumChannels+Channel]=Ptr[Cnt*NumChannels+Channel];
    }
  }
  
  /* clear everything else */
  for (Cnt=Size-OutputBufferSize;Cnt < Size;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      Ptr[Cnt*NumChannels+Channel]=0;
    }
  }

  return FCB;
}


void load_nop_filter(void)
{
  if(float_filter) free(float_filter);
  if (int_filter) free(int_filter);
  float_filter=malloc(sizeof(float));
  int_filter=malloc(sizeof(INT32));
  *float_filter=1.0;
  *int_filter=32767;
  filterlength=1;
}


void loadfilter(char *name)
{
  FILE *ffile;
  int Size;
  int Cnt;
  double Num;

  ffile=fopen(name,"r");
  if (!ffile) {
    fprintf(stderr,"Error opening filter file: %s\n",name);
    return;
  }
  /* first count size */
  for (Size=0;fscanf(ffile," %lg ",&Num) > 0;Size++);
  if (!Size) {
    fprintf(stderr,"Error loading filter file: %s\n",name);
    return;
  }
  rewind(ffile); /* restart reading file */

  /* now read the data */

  if (float_filter) free(float_filter);
  if (int_filter) free(int_filter);
  
  float_filter=malloc(Size*sizeof(float));
  int_filter=malloc(Size*sizeof(INT32));
  filterlength=Size;

  for (Cnt=0;(fscanf(ffile," %lg ",&Num) > 0) && Cnt < Size;Cnt++) {
    float_filter[Cnt]=Num;
    int_filter[Cnt]=(INT32)(Num*32767.0);
  }
  
  if (Cnt < Size) {
    fprintf(stderr,"Error reading filter file: %s\n",name);
    load_nop_filter();
    return;    
  }
  
  
  fclose(ffile);
}

int findnearestfactor(int number,int factor)
{
  int Cnt;
  for (Cnt=0;;Cnt++) {
    if ((number % (factor+ (Cnt/2)*( (Cnt & 1) * 2 - 1)))==0) {
      return (factor+ (Cnt/2)*( (Cnt & 1) * 2 - 1));
    }
  }
  return 1;
}

int main(int argc,char *argv[])
{
  int Cnt;
  struct SBuf *Buffer;
  struct SBuf *OutBuf;
  struct FilterControlBlock *FCB=NULL;
  struct queued_outbuffer *guilink=NULL;
  int newdownfactor;
  
  
  global_argv=argv;

  for (Cnt=1;argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    else if (!strncmp(argv[Cnt],"-downsamplefactor=",18)) {
      downsamplefactor=strtol(argv[Cnt]+18,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-filterfile=",12)) {
      filterfile=strdup(argv[Cnt]+12);
      
    } else if (!strcmp(argv[Cnt],"-nopanel")) {
      NoPanel=1;
    } else if (!strncmp(argv[Cnt],"-title=",7)) {
      WindowTitle=strdup(argv[Cnt]+7);
    }

  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   aglaophone format input\n"
	    "        #1   aglaophone format output\n"
	    "        #2   error messages (do not connect)\n"
      	    "\n"
	    "Usage:\n"
	    "   %s -magic -downsamplefactor=2 -filterfile=filename -filterdisabled -nopanel\n"
	    "         \"-title=Filter and downsample\"  \n"
	    "\n"
	    "   Defaults are upsamplefactor=2, filtering disabled.\n"
	    "   filterfile should be a text file with the filter taps\n"
	    ,argv[0]);
    exit(1);
  }
  
  if (!NoPanel) {
    initfdpi(argv[0],&guilink);
    if (!guilink) {
      fprintf(stderr,"%s: Error creating GUI\n",argv[0]);
      exit(1);
    }
  }
  
  Buffer=CreateSBuf();
  OutBuf=CreateSBuf();

  if (filterfile) {
    loadfilter(filterfile);
  } else {
    load_nop_filter();
  }

  for (;;) {
    if (!NoPanel) {
      /* handle GUI i/o */
      fdpi_readchars();
      if (queued_outbuffer_data_available(guilink)) write_queued_outbuffer(guilink);
    }
    
    /* Read in data from input */
    SBufPrepareHeader(Buffer);
    ReadSBuf(Buffer,readpipe);
    assert(SBufCheckHeader(Buffer));
    SBufReallocateIfNeeded(Buffer);
    SBufPrepareFinish(Buffer);
    ReadSBuf(Buffer,readpipe);
    
    if (Buffer->Header->BlockSize != BlockSize) {
      BlockSize=Buffer->Header->BlockSize;
    }
    if (Buffer->Header->NumChannels != NumChannels) {
      NumChannels=Buffer->Header->NumChannels;
    }

    if (Buffer->Header->SampleRate != SampleRate) {
      SampleRate=Buffer->Header->SampleRate;
    }
    
    
    SBufCopyHeader(Buffer,OutBuf);
    if (BlockSize % downsamplefactor) {
      newdownfactor=findnearestfactor(BlockSize,downsamplefactor);
      fprintf(stderr,"%s: Can't downsample by non-factor %d of the blocksize (bs=%d). Using %d\n",argv[0],downsamplefactor,BlockSize,newdownfactor);
      downsamplefactor=newdownfactor;
      if (!NoPanel) {
	fdpi_setdownsamplefactor(guilink,downsamplefactor);
      }
      assert((BlockSize % downsamplefactor)==0);
    }
    
    OutBuf->Header->SampleRate=SampleRate/downsamplefactor;
    OutBuf->Header->BlockSize=BlockSize/downsamplefactor;
    SBufReallocateIfNeeded(OutBuf);

    FCB=FilterAndDownsample(Buffer->Data,OutBuf->Data,downsamplefactor,NumChannels,Buffer->Header->BlockSize,FCB,int_filter,filterlength);    
    
    /* Write data to output */
    SBufPrepare(OutBuf);
    WriteSBuf(OutBuf,writepipe);
    
    
  }

  
}

