/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; 
     version 2.1 of the License.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

*/

#include <sys/types.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "io_spec.h"
#include "buffer.h"

int SBufIsDone(struct SBuf *Buf)
{
  if (Buf->Bytes && Buf->Pos==Buf->Bytes)
    return 1;
  else return 0;
}

int SBufIsHeaderDone(struct SBuf *Buf)
{
  if (Buf->Bytes==sizeof(struct AglBlk) && Buf->Pos==Buf->Bytes)
    return 1;
  else return 0;
  
}

int SBufIsDataDone(struct SBuf *Buf)
{
  if (Buf->Bytes > sizeof(struct AglBlk) && Buf->Pos==Buf->Bytes)
    return 1;
  else return 0;
  
}


void SBufNewSize(struct SBuf *Buf,int NewSize)
     /* Reallocates Buf->Data, sets buffer as empty  */
{
  if (Buf->Header) free(Buf->Header);

  Buf->Size=NewSize;
  Buf->Header=malloc(sizeof(struct AglBlk)+Buf->Size*Buf->NumChannels*sizeof(INT16));
  Buf->Data=(INT16 *)(Buf->Header+1);
  Buf->Bytes=0;
  Buf->Pos=0;
  
  
}

int SBufIsReady(struct SBuf *Buf)
{
  return (Buf->Bytes > 0 && Buf->Pos < Buf->Bytes);
}

void SBufPrepareHeader(struct SBuf *Buf)
     /* Prepare to send/receive a header only */
{
  Buf->Bytes=sizeof(struct AglBlk);
  Buf->Pos=0;
}

void SBufPrepareFinish(struct SBuf *Buf)
     /* Prepare to send/receive data, given that the header has
      already been received/sent*/
{
  Buf->Bytes=sizeof(struct AglBlk)+Buf->Size*Buf->NumChannels*sizeof(INT16);
  Buf->Pos=sizeof(struct AglBlk);
}

void SBufPrepare(struct SBuf *Buf)
     /* Prepare to send/receive a full block */
     /* This can be used to send a full block. It should generally not
	be used to receive, because it assumes that the block size will 
	not change. The proper sequence for receiving is
	SBufPrepareHeader();
	SBufRead();
	assert(SBufCheckHeader())
	SBufReallocateIfNeeded();
	SBufPrepareFinish();
	SBufRead() */
     
{
  Buf->Bytes=sizeof(struct AglBlk)+Buf->Size*Buf->NumChannels*sizeof(INT16);
  Buf->Pos=0;
}


void SBufSetHeader(struct SBuf *Buf,int NumChannels,int SampleRate,int FrameNumber, int BlockSize)
{
  struct AglBlk *H;
  
  H=Buf->Header;
  H->MagicByte[0]=H->MagicByte[1]=H->MagicByte[2]=H->MagicByte[3]=H->MagicByte[4]=H->MagicByte[5]=0x7f;
  H->zero=0;
  H->NumChannels=NumChannels;
  H->SampleRate=SampleRate;
  H->FrameNumber=FrameNumber;
  H->BlockSize=BlockSize;

  H->MoreMagicByte[0]=H->MoreMagicByte[1]=H->MoreMagicByte[2]=0x7f;
  H->One=1;
  H->Reserved=0;
}

struct AglBlk *SBufGetHeader(struct SBuf *Buf)
     /* The returned pointer is into the buffer data. If this 
	gets reallocated or overwritten the returned pointer is
	no longer valid! */
{
  return Buf->Header;
}

int SBufCopyHeader(struct SBuf *Source, struct SBuf *Dest)
     /* Copy the data header from Source to Dest. Returns non-zero if 
	header data requires reallocation (i.e. if Size*NumChannels
	has changed) */
{
  struct AglBlk *HS,*HD;
  int Change;

  HS=Source->Header;
  HD=Dest->Header;

  Change=((HS->BlockSize*HS->NumChannels) != (HD->BlockSize * HD->NumChannels));
  *HD=*HS;
  return Change;
}





struct SBuf *CreateSBuf(void)
     /* Reserves enough data for the header only */
{
  struct SBuf *Blk;
  
  Blk=malloc(sizeof(struct SBuf));
  Blk->Header=malloc(sizeof(struct AglBlk));
  Blk->Data=NULL;
  Blk->Size=0;
  Blk->NumChannels=0;
  Blk->Bytes=0;
  Blk->Pos=0;
  return Blk;
}

void DeleteSBuf(struct SBuf *Buf)
     /* Free resources associated with an SBuf */
{
  if (Buf->Header) free(Buf->Header);
  free(Buf);
  
}

int WriteSBuf(struct SBuf *Buf,int Pipe)
     /* returns 1 when the SBuf has been fully written,
	0 otherwise */
{
  int NumBytes;

  while (Buf->Pos < Buf->Bytes) {
    NumBytes=write(Pipe, ((char *)Buf->Header)+Buf->Pos,Buf->Bytes-Buf->Pos);
    
    if (NumBytes < 0 && errno==EWOULDBLOCK) {
      /* can't write -- return !! */
      return 0;
    } else if (NumBytes < 0 && errno != EINTR && errno != EAGAIN) {
      /* Error !!! */
      perror("WriteSBuf");
      exit(1);
    } else if (NumBytes > 0)
      Buf->Pos+=NumBytes;

  }
  
  return 1;

  
}




int ReadSBuf(struct SBuf *Buf,int Pipe)
     /* returns 1 when the SBuf has been fully read,
	0 otherwise */
{
  int NumBytes;

  while (Buf->Pos < Buf->Bytes) {
    NumBytes=read(Pipe, ((char *)Buf->Header)+Buf->Pos,Buf->Bytes-Buf->Pos);

    if (NumBytes < 0 && errno==EWOULDBLOCK) {
      /* can't read -- return !! */
      return 0;
    } else if (NumBytes < 0 && errno != EINTR && errno != EAGAIN) {
      /* Error !!! */
      perror("ReadSBuf");
      exit(1);
    } else if (NumBytes==0) {
      /* EOF */
      exit(0);
    } else if (NumBytes > 0)
      Buf->Pos+=NumBytes;

  }
  
  return 1;

  
}




int SBufReallocateIfNeeded(struct SBuf *Buf)
/* SBufReallocateIfNeeded() looks at the StatusBlk in Buf,
   determines if reallocation of the data area is necessary,
   and does so. Be warned that the status block will no longer
   be at the same location after this call. Returns 1 if it 
   performed reallocation.
*/
{
  struct AglBlk *Stat;
  
  Stat=Buf->Header;
  
  if (Stat->NumChannels * Stat->BlockSize != Buf->Size * Buf->NumChannels) {
    Buf->NumChannels=Stat->NumChannels;
    Buf->Size=Stat->BlockSize;
    
    Buf->Header=malloc(Buf->Size*Buf->NumChannels*sizeof(INT16)+sizeof(struct AglBlk));
    Buf->Data=(INT16 *)(Buf->Header+1);
    *(Buf->Header)=*Stat;
    Buf->Bytes=sizeof(struct AglBlk)+Buf->Size*Buf->NumChannels*sizeof(INT16);
    Buf->Pos=0;
    free(Stat); /* free old data buffer */
    
    return 1;
  } else if (Stat->NumChannels != Buf->NumChannels || Stat->BlockSize != Buf->Size) {
    /* change in number of channels that cancels change in block size */
    Buf->Size=Stat->BlockSize;
    Buf->NumChannels=Stat->NumChannels;
      
  }
  return 0;
}

int SBufCheckHeader(struct SBuf *Buf)
     /* do a sanity check on the header. Return 
	non-zero if OK, 0 otherwise */
{
  char *CmpBuf="\x7f\x7f\x7f\x7f\x7f\x7f";
  
  if (memcmp(&Buf->Header->MagicByte[0],CmpBuf,6)) return 0;
  
  if (Buf->Header->zero != 0) return 0;
  if (Buf->Header->NumChannels > 64) return 0; /* nonsensical number of channels */
  if (memcmp(&Buf->Header->MoreMagicByte[0],CmpBuf,3)) return 0;
  if (Buf->Header->One != 1) return 0;
  return 1;
}

int SBufIsEmpty(struct SBuf *Buf)
{
  if (Buf->Bytes==0 && Buf->Pos==0) return 1;
  return 0;
}

void SBufEmpty(struct SBuf *Buf)
{
  Buf->Bytes=0;
  Buf->Pos=0;
}
