/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; 
     version 2.1 of the License.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA


*/

#include <sys/types.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "io_spec_old.h"
#include "oldbuffer.h"

int SBufIsDone(struct SBuf *Buf)
{
  if (Buf->Bytes && Buf->Pos==Buf->Bytes)
    return 1;
  else return 0;
}

void SBufNewSize(struct SBuf *Buf,int NewSize)
     /* Reallocates Buf->Data, clears buffer.  */
{
  if (Buf->Data) free(Buf->Data);

  Buf->Size=NewSize;
  Buf->Data=malloc(Buf->Size*Buf->NumChannels*sizeof(INT16)+sizeof(struct DataBlkHeader));
  Buf->Bytes=0;
  Buf->Pos=0;
  
  
}

int SBufIsReady(struct SBuf *Buf)
{
  return (Buf->Bytes > 0 && Buf->Pos < Buf->Bytes);
}


void SBufPrepare(struct SBuf *Buf)
     /* Prepare to send/receive a full block */
{
  Buf->Bytes=Buf->Size*Buf->NumChannels*sizeof(INT16) + sizeof(struct DataBlkHeader);
  Buf->Pos=0;
}

void SBufPrepareStatus(struct SBuf *Buf)
     /* Prepare to send/receive a status block */
{
  Buf->Bytes=sizeof(struct StatusBlk);
  Buf->Pos=0;
}     

void SBufSetDataHeader(struct SBuf *Buf,int NextIsStatusBlk)
{
  struct DataBlkHeader *H;
  
  H=(struct DataBlkHeader *)Buf->Data;
  H->MagicByte[0]=H->MagicByte[1]=H->MagicByte[2]=0x7f;
  H->NextBlockIsStatusBlk=NextIsStatusBlk;
}

struct DataBlkHeader *SBufGetDataHeader(struct SBuf *Buf)
     /* The returned pointer is into the buffer data. If this 
	gets reallocated or overwritten the returned pointer is
	no longer valid! */
{
  return (struct DataBlkHeader *)Buf->Data;
}

int SBufCopyDataHeader(struct SBuf *Source, struct SBuf *Dest)
     /* Copy the data header from Source to Dest. Return 
	NextBlockIsStatusBlk field */
{
  struct DataBlkHeader *HS,*HD;
  
  HS=(struct DataBlkHeader *)Source->Data;
  HD=(struct DataBlkHeader *)Dest->Data;
  *HD=*HS;
  return HD->NextBlockIsStatusBlk;
}


void SBufSetStatusBlk(struct SBuf *Buf,int NumChannels,int SampleRate,int FrameNumber,int BlockSize)
{
  struct StatusBlk *Blk;

  Blk=(struct StatusBlk *)Buf->Data;
  Blk->MagicByte[0]=Blk->MagicByte[1]=Blk->MagicByte[2]=Blk->MagicByte[3]=Blk->MagicByte[4]=Blk->MagicByte[5]=0x7f;
  Blk->zero=0;
  Blk->NumChannels=NumChannels;
  Blk->SampleRate=SampleRate;
  Blk->FrameNumber=FrameNumber;
  Blk->BlockSize=BlockSize;
  Buf->Bytes=sizeof(struct StatusBlk);
  Buf->Pos=0;
}


struct StatusBlk *SBufGetStatusBlk(struct SBuf *Buf)
{
  return (struct StatusBlk *)Buf->Data;
}


int SBufCopyStatusBlk(struct SBuf *Src,struct SBuf *Dest)
     /* Returns 1 if the status block data requires a newly 
	reallocated block */
{
  struct StatusBlk *SS,*SD;
  SS=(struct StatusBlk *)Src->Data;
  SD=(struct StatusBlk *)Dest->Data;

  (*SD)=(*SS);
  
  if (SD->NumChannels * SD->BlockSize != Dest->Size * Dest->NumChannels) {
    return 1;
  }
  return 0;
}

struct SBuf *CreateSBuf(void)
     /* Reserves enough data for a StatusBlk */
{
  struct SBuf *Blk;
  
  Blk=malloc(sizeof(struct SBuf));
  Blk->Data=malloc(sizeof(struct StatusBlk));
  Blk->Size=0;
  Blk->NumChannels=0;
  Blk->Bytes=0;
  Blk->Pos=0;
  return Blk;
}

void DeleteSBuf(struct SBuf *Buf)
     /* Free resources associated with an SBuf */
{
  if (Buf->Data) free(Buf->Data);
  free(Buf);
  
}

int WriteSBuf(struct SBuf *Buf,int Pipe)
     /* returns 1 when the SBuf has been fully written,
	0 otherwise */
{
  int NumBytes;

  while (Buf->Pos < Buf->Bytes) {
    NumBytes=write(Pipe, ((char *)Buf->Data)+Buf->Pos,Buf->Bytes-Buf->Pos);

    if (NumBytes < 0 && errno==EWOULDBLOCK) {
      /* can't write -- return !! */
      return 0;
    } else if (NumBytes < 0 && errno != EINTR && errno != EAGAIN) {
      /* Error !!! */
      perror("WriteSBuf");
      exit(1);
    } else if (NumBytes > 0)
      Buf->Pos+=NumBytes;

  }
  
  return 1;

  
}


int ReadSBuf(struct SBuf *Buf,int Pipe)
     /* returns 1 when the SBuf has been fully read,
	0 otherwise */
{
  int NumBytes;

  while (Buf->Pos < Buf->Bytes) {
    NumBytes=read(Pipe, ((char *)Buf->Data)+Buf->Pos,Buf->Bytes-Buf->Pos);

    if (NumBytes < 0 && errno==EWOULDBLOCK) {
      /* can't read -- return !! */
      return 0;
    } else if (NumBytes < 0 && errno != EINTR && errno != EAGAIN) {
      /* Error !!! */
      perror("ReadSBuf");
      exit(1);
    } else if (NumBytes==0) {
      /* EOF */
      exit(0);
    } else if (NumBytes > 0)
      Buf->Pos+=NumBytes;

  }
  
  return 1;

  
}


int SBufIsStatus(struct SBuf *Buf)
 /* returns 1 if this is a status block */
{
  struct StatusBlk *Stat;

  Stat=(struct StatusBlk *)Buf->Data;
  if (Stat->MagicByte[3]==0x7f)
    return 1;
  else return 0;

}

struct StatusBlk *SBufGetStatus(struct SBuf *Buf)
{
  return (struct StatusBlk *)Buf->Data;
}

int SBufReallocateIfNeeded(struct SBuf *Buf)
/* SBufReallocateIfNeeded() looks at the StatusBlk in Buf,
   determines if reallocation of the data area is necessary,
   and does so. Be warned that the status block will no longer
   be at the same location after this call. Returns 1 if it 
   performed reallocation.
*/
{
  struct StatusBlk *Stat;
  
  Stat=(struct StatusBlk *)Buf->Data;
  
  if (Stat->NumChannels * Stat->BlockSize != Buf->Size * Buf->NumChannels) {
    Buf->NumChannels=Stat->NumChannels;
    Buf->Size=Stat->BlockSize;
    
    Buf->Data=malloc(Buf->Size*Buf->NumChannels*sizeof(INT16)+sizeof(struct DataBlkHeader));
    memcpy(Buf->Data,Stat,sizeof(struct StatusBlk));
    Buf->Bytes=sizeof(struct StatusBlk);
    free(Stat); /* free old data buffer */
    
    return 1;
  }
  return 0;
}

int SBufCheckStatus(struct SBuf *Buf)
     /* do a sanity check on the status block. Return 
	non-zero if OK, 0 otherwise */
{
  char *CmpBuf="\x7f\x7f\x7f\x7f\x7f\x7f";
  struct StatusBlk *Stat;
  
  Stat=SBufGetStatusBlk(Buf);
  if (memcmp(&Stat->MagicByte[0],CmpBuf,6)) return 0;
  
  if (Stat->zero != 0) return 0;
  if (Stat->NumChannels > 64) return 0; /* nonsensical number of channels */
  return 1;
}

int SBufIsEmpty(struct SBuf *Buf)
{
  if (Buf->Bytes==0 && Buf->Pos==0) return 1;
  return 0;
}
