/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
     
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; 
     version 2.1 of the License.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

*/

#include <stdarg.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <stdlib.h>
#include <errno.h>

#include "linklist.h"
#include "util.h"

struct timeval starttime;

void startdebugtimer(void)
     /* warning: the debugtimer is NON-REENTRANT */
{
  gettimeofday(&starttime,NULL);
  
}

float resultdebugtimer(void)
{
  struct timeval endtime;
  int deltasec;
  int deltausec;
  float deltafloat;

  gettimeofday(&endtime,NULL);
  
  deltasec=endtime.tv_sec-starttime.tv_sec;
  deltausec=endtime.tv_usec-starttime.tv_usec;
  /*  fprintf(stderr,"deltausec=%ld\n",(long)deltausec);*/
  if (deltausec < 0) {
    deltausec+=1000000;
    deltasec--;
  }
  deltafloat=(float)deltasec+1.e-6*deltausec;
  
}

void setnonblocking(int fd)
{
  /* This may need to be changed for different versions of UNIX */
  fcntl(fd, F_SETFL,O_NONBLOCK);
  
}

int maxva(int arg,...)
     /* find the largest value in a variable-length argument list 
	terminated by -2. */
{
  int largest=-2;
  int curarg;
  
  va_list ap;
  
  va_start(ap,arg);
  
  largest=arg;

  do {
    curarg=va_arg(ap,int);
    if (curarg > largest) largest=curarg;
  } while (curarg != -2);

  va_end(ap);
  return largest;
}


char *search_argv0(char *argv0)
     /* search the command path to find the actual (?) path
	to the command which was run. We do this by opening a 
	pipe to which(1). Be warned that there are race conditions
	by which this might give the path to a different executable
	of the same name (!). But there doesn't seem to be a better way. 
	This routine DOES NOT trace back symbolic links. */
{
  /* open a pipe to which(1) */
  int childid;
  int pipefds[2];
  FILE *Readfile;
  char PathBuf[800];
  struct stat buf;
  
  if (!stat(argv0,&buf)) return strdup(argv0);

  pipe(pipefds);
  
  if (!(childid=fork())) {
    /* child process */
    close(pipefds[0]);
    dup2(pipefds[1],1);
    
    execlp("which","which",argv0,NULL);
    exit(1);
  }
  close(pipefds[1]);
  if (childid < 0) {
    /* some sort of error */
    close(pipefds[0]);
    return strdup(argv0);
  }
  Readfile=fdopen(pipefds[0],"r");
  PathBuf[0]=0;
  fgets(PathBuf,799,Readfile);

  waitpid(childid,NULL,0);
  fclose(Readfile);
  
  /* remove trailing newlines */
  while (strlen(PathBuf) && isspace(PathBuf[strlen(PathBuf)-1]))
    PathBuf[strlen(PathBuf)-1]='\0';

  if (!strlen(PathBuf)) return strdup(argv0);
  
  return strdup(PathBuf);
}

char *find_argv0(char *argv0)
     /* search the command path to find the actual (?) path
	to the command which was run. We do this by opening a 
	pipe to which(1). Be warned that there are race conditions
	by which this might give the path to a different executable
	of the same name (!). But there doesn't seem to be a better way. 
	This routine DOES trace back symbolic links. */
{
  /* first search the path, then traverse any symbolic links */
  char pathbuf[MAXPATHLEN];

  char *newpath;

  newpath=search_argv0(argv0);
  
  if (realpath(newpath,pathbuf)) {
    free(newpath);
    /* fprintf(stderr,"%s\n",pathbuf); */
    return strdup(pathbuf);
  }
  else {
    
    /* fprintf(stderr,"%s\n",newpath); */
    return newpath;
    
  }

}

struct queued_outbuffer *create_queued_outbuffer(int fd)
{
  struct queued_outbuffer *ret;
  
  ret=malloc(sizeof(struct queued_outbuffer));
  
  NewList(&ret->queue);
  ret->writefd=fd;
  
  return ret;
}

void delete_queued_outbuffer(struct queued_outbuffer *queue)
     /* does not close file descriptor */
{
  struct queued_elem *elem;
  while (elem=(struct queued_elem *)RemHead(&queue->queue)) free(elem);
  free(queue);
  
}

int queued_outbuffer_data_available(struct queued_outbuffer *buf)
{
  return !!(buf->queue.lh_Head->ln_Succ);
}

void write_queued_outbuffer(struct queued_outbuffer *buf)
{
  struct queued_elem *elem;
  int Ret;

  while ((elem=(struct queued_elem *)buf->queue.lh_Head)->Node.ln_Succ) { /* While there are any elements on the queue */
    Ret=write(buf->writefd,elem->Buffer+elem->BufPos,elem->BufLen-elem->BufPos);
    if (Ret < 0) {
      if (errno==EINTR) continue;
      else if (errno==EAGAIN || errno==EWOULDBLOCK) return; /* pipe cannot accept data */
      else {
	perror("Write error on queued buffer");
	exit(1);
      }
    } else if (Ret==0) {
      /* EOF -- user closed window -- exit silently */
      exit(0);
    } else {
      elem->BufPos+=Ret;
      if (elem->BufPos>=elem->BufLen) {
	/* Buffer completely sent -- dequeue it */
	Remove((struct Node *)elem);
	free(elem);
      }
    }
  }
}

void queue_outbuffer(struct queued_outbuffer *queue,int priority,char *Data,int Len) 
     /* priority of 0 means don't queue anything if there's more than one element
	already queued. Priority 1 means always queue */
{
  struct queued_elem *elem;

  if (priority==0 && queue->queue.lh_Head->ln_Succ && queue->queue.lh_Head->ln_Succ->ln_Succ)
    return; /* don't queue */

  if (Len==0) return; /* can't queue an empty buffer! */
  
  elem=malloc(sizeof(struct queued_elem)+Len);
  elem->Buffer=(char *)(elem+1);
  elem->BufLen=Len;
  elem->BufPos=0;
  memcpy(elem->Buffer,Data,Len);
  AddTail(&queue->queue,(struct Node *)elem);
}

void qob_printf(struct queued_outbuffer *queue,int priority,char *fmt,...) /* printf to a queued output buffer */
{
  va_list ap;
  int Ret=0;
  static char *Buf;
  static int BufSize;
  
  va_start(ap,fmt);
  
  if (!BufSize) {
    BufSize=200;
    Buf=malloc(BufSize);
  }
  
  
  do {
    if (Ret) {
      BufSize <<= 1;
      free(Buf);
      Buf=malloc(BufSize);
    }
    Ret=vsnprintf(Buf,BufSize,fmt,ap);
  } while (Ret==BufSize);
  
  queue_outbuffer(queue,priority,Buf,strlen(Buf));
  
  va_end(ap);
}


#define RDBUFSIZE 200

void parserreadchars(int readfd,int (*parsefunc)(int Code),void (*initfunc)(void))
     /* parsefunc takes the character Code, runs it through the parser, 
	and returns 1 if the parse is continuing, or zero if the parse needs to 
	be restarted (by calling initfunc()) */
{

  int GotBytes;
  char Buf[RDBUFSIZE];
  int Cnt;
  
  
  do {
    GotBytes=read(readfd,Buf,RDBUFSIZE);
    if (!GotBytes) {
      /* EOF on connection. User has closed us, so we die. */
      exit(0);
    } else if (GotBytes < 0) {
      if (errno==EINTR) continue;
      else if (errno==EWOULDBLOCK || errno==EAGAIN) return; /* no data available */
      else {
	perror("util/parserreadchars/read()");
	exit(1);
      }
    } else {
      for (Cnt=0;Cnt < GotBytes;Cnt++) {
	/*	  fprintf(stderr,"calling parser with character %c.\n",PCB.input_code);*/
	if (!((*parsefunc)(Buf[Cnt]))) {
	  fprintf(stderr,"util/parserreadchars: failed due to parse error. Parser restarted.\n");
	  (*initfunc)(); /* restart parser */
	}
      }
    }
    
  } while (GotBytes > 0);
  
}

void spawnjava(char *argv0,char *classname,int *writetojava,int *readfromjava)
{
  char *execpath;
  int childinputfds[2];
  int childoutputfds[2];
  int childid;
  char *newclasspath;
  char *classpath;
  char *javapath1;
  char *javapath2;
  
  int junk;

  if (!writetojava) writetojava=&junk;
  if (!readfromjava) readfromjava=&junk;
  


  classpath=getenv("CLASSPATH");
  if (!classpath) classpath="";

  execpath=find_argv0(argv0); /* determine path to our executable */

  if (!execpath || !strlen(execpath)) {
    execpath=malloc(3);
    strcpy(execpath,"./");
  }
  
  /* strip filename off execpath */
  while (strlen(execpath) && execpath[strlen(execpath)-1]!='/')
    execpath[strlen(execpath)-1]=0;
  if (!strlen(execpath)) strcpy(execpath,"./");

  newclasspath=malloc(strlen(classpath)+strlen(execpath)+5);
  strcpy(newclasspath,execpath);
  strcat(newclasspath,":");
  strcat(newclasspath,classpath);

  javapath1=malloc(strlen(execpath)+strlen(JAVAPATH)+10);
  sprintf(javapath1,"%s/../%s",execpath,JAVAPATH);

  javapath2=malloc(strlen(execpath)+strlen(JAVAPATH)+10);
  sprintf(javapath2,"%s/%s",execpath,JAVAPATH);
  

  /* should add jre path specification sometime */
  
  /* spawn off process */
  pipe(childinputfds);
  pipe(childoutputfds);

  /*fprintf(stderr,"classpath=%s\n",newclasspath);*/
  
  if (!(childid=fork())) {
    /* child process */
    close(childoutputfds[0]);
    close(childinputfds[1]);
    dup2(childoutputfds[1],1);
    dup2(childinputfds[0],0);
    
    /* if we obtained a special priority due to 
       root status, relinquish thie priority for the 
       java GUI */
    setpriority(PRIO_PROCESS,0,0);

    /* Work around JRE bug */
    putenv("LD_ASSUME_KERNEL=2.2.5");

    /* try various places to find the java executable */

    execlp(javapath1,javapath1,"-classpath",newclasspath,classname,NULL);
    execlp(javapath2,javapath2,"-classpath",newclasspath,classname,NULL);
    execlp(JAVAPATH,JAVAPATH,"-classpath",newclasspath,classname,NULL);
    execlp("java","java","-classpath",newclasspath,classname,NULL);

    perror("soundcardio/panelinterface: Attempting to run java (Java Runtime Environment)");
    exit(1);
  }
  close(childoutputfds[1]);
  close(childinputfds[0]);
  
  *readfromjava=childoutputfds[0];
  *writetojava=childinputfds[1];
  
}

