/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* pacer.c -- Take an input stream which is not rate limited
              and limit it according to the sample rate specified
	      in the stream */

/* Known bug: This module is highly dependent upon the system clock. If the
   system clock does not run at the correct rate, results may be poor. 
   
   Furthermore, expect this module to stop working correctly if the system
   time is changed while it is running. */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "io_spec_old.h"
#include "oldbuffer.h"



struct SBuf *Buf;

int InputPipe=0;
int OutputPipe=1;

int SampleRate;
int BlkSize;


int main(int argc,char *argv[])
{
  int Cnt;
  int GotMagic=0;
  signed long usecdiff;
  struct StatusBlk *Status;
  long Delay;
  long DelaySec;
  
  struct timeval targettime,curtime,waittime;
  
  for (Cnt=1; Cnt < argc; Cnt++) {
    if (!strncmp(argv[Cnt],"-magic",6)) {
      GotMagic=1;
    }
  }
  
  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:   #0     Non-rate-limited stream input\n"
	    "         #1     Rate-limited stream output\n"
	    "         #2     Error messages (do not connect)\n"
	    "\n"
	    "Usage;\n"
	    "   %s -magic\n"
	    "\n"
	    "NOTE: This module is highly dependent upon the system clock. If the\n"
	    "system clock does not run at the correct rate, results may be poor.\n" 
	    "\n"
	    "Furthermore, expect this module to stop working correctly if the system\n"
	    "time is changed while it is running.\n"
	    ,argv[0]);
    exit(1);
  }

  /* Set up buffer */
  Buf=CreateSBuf();  
  
  /* The first block to read must be a status block */
  SBufPrepareStatus(Buf);
  
  /* we want to send the first data block immediately. */
  gettimeofday(&targettime,NULL);
  
  
  for(;;) {
    /* begin reading and writing */
    while(!ReadSBuf(Buf,InputPipe));

    if (SBufIsStatus(Buf)) {
      Status=SBufGetStatus(Buf);
      SampleRate=Status->SampleRate;
      BlkSize=Status->BlockSize;
      
      /* If we needed to change it, we could do so here */

      /* Send out the status block. */
      SBufPrepareStatus(Buf);
      while(!WriteSBuf(Buf,OutputPipe));
      
      SBufReallocateIfNeeded(Buf);
      
      /* Then get the next data block */
      SBufPrepare(Buf);
      while(!ReadSBuf(Buf,InputPipe));
    }
    /* We have a data block. If we wanted to do in-place processing
       we could do that here */

    /* wait until targettime */
    for (gettimeofday(&curtime,NULL); timercmp(&curtime,&targettime, < ) ;gettimeofday(&curtime,NULL)) {
      waittime.tv_sec=targettime.tv_sec-curtime.tv_sec;
      usecdiff=targettime.tv_usec-curtime.tv_usec;
      if (usecdiff < 0) {
	waittime.tv_sec--;
	usecdiff+=1000000;
      }
      waittime.tv_usec=usecdiff;
      
      /* perform the wait with select() */
      select(0,NULL,NULL,NULL,&waittime);
    }
    
    /* we've waited until the correct time */
    /* Select the appropriate delay for the next packet */
    Delay=(((long long)BlkSize)*((long long)1000000))/((long long)SampleRate); /* This is the delay, in us */
    DelaySec = Delay / 1000000;
    Delay -= DelaySec*1000000;
    
    targettime.tv_sec+=DelaySec;
    targettime.tv_usec+=Delay;
    if (targettime.tv_usec > 1000000) {
      targettime.tv_sec++;
      targettime.tv_usec-=1000000;
    }
    
    /* and write out this buffer */
    SBufPrepare(Buf);
    while(!WriteSBuf(Buf,OutputPipe));
    
    
    /* go on and read the next buffer */
    if ( (SBufGetDataHeader(Buf))->NextBlockIsStatusBlk) 
      SBufPrepareStatus(Buf);
    else
      SBufPrepare(Buf);
  }
  
  
}



