/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "io_spec_old.h"
#include "filterdata.h"
#include "filter.h"


INT16 *upintfilter; /* this is an array of  values in the filterdata[]
		     array. They are integers, so we can use integer
		     multiplication instead of floating point multiplication */

int maxupfilterlength;


INT16 *linearupintfilter;
int maxlinearupfilterlength;

INT16 *downintfilter;
int maxdownfilterlength;


void BuildIntFilter(void)
{
  int Cnt;
  float num;

  maxupfilterlength=upfilterlengths[numfilters-1];
  upintfilter=malloc(numfilters*maxupfilterlength*sizeof(INT32));
  
  for (Cnt=0;Cnt < numfilters*maxupfilterlength; Cnt++) {
    num=upfilterdata[Cnt];
    assert(num >= -1.0 && num <= 1.0);
    
    upintfilter[Cnt]=num*32767.0;
  }

  maxlinearupfilterlength=linearupfilterlengths[numfilters-1];
  linearupintfilter=malloc(numfilters*maxlinearupfilterlength*sizeof(INT32));
  
  for (Cnt=0;Cnt < numfilters*maxlinearupfilterlength; Cnt++) {
    num=linearupfilterdata[Cnt];
    assert(num >= -1.0 && num <= 1.0);
    
    linearupintfilter[Cnt]=num*32767.0;
  }

  
  maxdownfilterlength=downfilterlengths[numfilters-1];
  downintfilter=malloc(numfilters*maxdownfilterlength*sizeof(INT32));
  
  for (Cnt=0;Cnt < numfilters*maxdownfilterlength; Cnt++) {
    num=downfilterdata[Cnt];
    assert(num >= -1.0 && num <= 1.0);
    
    downintfilter[Cnt]=num*32767.0;
  }
  
}


struct FilterControlBlock *FilterAndUpsample(INT16 *InputBuf,INT16 *OutputBuf,int UpsampleAmnt,int NumChannels,int OutputBufferSize,struct FilterControlBlock *FCB,int linear)
     /* use linear interpolation instead of regular filtering if linear is 
	set to 1 */
{
  int filternum,filterlength;
  int Size;
  int Cnt,Cnt2,Channel;
  INT32 *Ptr;
  int InputBufferSize;
  INT16 Value;
  INT16 *filter;
  
  if (!upintfilter || !linearupintfilter) BuildIntFilter();

  assert((OutputBufferSize % UpsampleAmnt)==0);
  InputBufferSize=OutputBufferSize / UpsampleAmnt;

  if (UpsampleAmnt==1) {
    /* just copy */

    memcpy(OutputBuf,InputBuf,OutputBufferSize*NumChannels*sizeof(INT16));
    
    return FCB;
  }

  if (!FCB) {
    FCB=malloc(sizeof(struct FilterControlBlock));
    FCB->SampleFactor=UpsampleAmnt;
    FCB->NumChannels=NumChannels;
    FCB->WorkBuf=NULL;
    FCB->WorkBufSize=0;
  }

  if (FCB->SampleFactor != UpsampleAmnt || FCB->NumChannels != NumChannels) FCB->WorkBufSize=0;
  
  FCB->SampleFactor=UpsampleAmnt;
  FCB->NumChannels=NumChannels;

  filternum=UpsampleAmnt-2;
  if (filternum >= numfilters) {
    fprintf(stderr,"Can't filter for an UpsampleAmnt of %d\n",UpsampleAmnt);
    exit(1);
  }
  /* fprintf(stderr,"Using filter number %d\n",(int)filternum); */
  if (linear) {
    filterlength=linearupfilterlengths[filternum];
    filter=&linearupintfilter[filternum*maxlinearupfilterlength];
  }
  else {
    filterlength=upfilterlengths[filternum];
    filter=&upintfilter[filternum*maxupfilterlength];
  }

  Size=OutputBufferSize+filterlength-1;
  if (FCB->WorkBufSize != Size) {
    if (FCB->WorkBuf) free(FCB->WorkBuf);
    FCB->WorkBufSize=Size;
    /* allocate and fill with zeros */
    FCB->WorkBuf=calloc(FCB->WorkBufSize*FCB->NumChannels*sizeof(INT32),1);
  }
  
  /* Now perform the filtering */
  /* we use a time-domain algorithm */
  /* This is order (filterlength * inputlength) per channel */
  /* relative to the nominal (downsampled) filter length and nominal (upsampled) sample rate,
     it is also (filterlength*inputlength). As of 1/19/99, nominal filterlength is 11 samples,
     so performance is O(11n) where n is the upsampled data length. */
  Ptr=FCB->WorkBuf;
  

  for (Cnt=0;Cnt < InputBufferSize;Cnt++) {

    for (Channel=0;Channel < NumChannels;Channel++) {
      /* on upsampling only, we scale by the scalefactor specified 
	 in the filterdata */
      Value=InputBuf[Cnt*NumChannels + Channel];/*/scalefactors[filternum];*/
      for (Cnt2=0;Cnt2 < filterlength;Cnt2++) {
	Ptr[(Cnt*UpsampleAmnt+Cnt2)*NumChannels+Channel]+=Value*filter[Cnt2];
      }
      
    }
  }

  /* done the filtering. Now extract the filtered data into the outputbuf, and move the remains
     into the start of the WorkBuf, where it will be used by the next filtering operation */

  /* copy to output buf */
  for (Cnt=0; Cnt < OutputBufferSize;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      OutputBuf[Cnt*NumChannels+Channel]=(Ptr[Cnt*NumChannels+Channel] >> 15); /* Take only high 
										  order 17 bits */
    }
  }
  
  /* move tail end */
  for (Cnt=OutputBufferSize; Cnt < FCB->WorkBufSize;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      Ptr[(Cnt-OutputBufferSize)*NumChannels+Channel]=Ptr[Cnt*NumChannels+Channel];
    }
  }
  
  /* clear everything else */
  for (Cnt=FCB->WorkBufSize-OutputBufferSize;Cnt < FCB->WorkBufSize;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      Ptr[Cnt*NumChannels+Channel]=0;
    }
  }

  return FCB;
}


struct FilterControlBlock *FilterAndDownsample(INT16 *InputBuf,INT16 *OutputBuf,int DownsampleAmnt,int NumChannels,int InputBufferSize,struct FilterControlBlock *FCB)
{
  int OutputBufferSize;
  int filternum;
  int filterlength;
  INT16 *filter;
  int Size;
  int Cnt,Cnt2,Channel;
  INT16 Value;
  INT32 *Ptr;

  if (!downintfilter) BuildIntFilter();

  if (DownsampleAmnt==1) {
    /* just copy */

    memcpy(OutputBuf,InputBuf,InputBufferSize*NumChannels*sizeof(INT16));
    
    return FCB;
  }

  assert((InputBufferSize % DownsampleAmnt)==0);
  OutputBufferSize=InputBufferSize/DownsampleAmnt;


  if (!FCB) {
    FCB=malloc(sizeof(struct FilterControlBlock));
    FCB->SampleFactor=DownsampleAmnt;
    FCB->NumChannels=NumChannels;
    FCB->WorkBuf=NULL;
    FCB->WorkBufSize=0;
  }
  
  

  if (FCB->SampleFactor != DownsampleAmnt || FCB->NumChannels != NumChannels) FCB->WorkBufSize=0;
  
  FCB->SampleFactor=DownsampleAmnt;
  FCB->NumChannels=NumChannels;

  filternum=DownsampleAmnt-2;

  if (filternum >= numfilters) {
    fprintf(stderr,"Can't filter for an DownsampleAmnt of %d\n",DownsampleAmnt);
    exit(1);
  }

  filterlength=downfilterlengths[filternum];
  filter=&downintfilter[filternum*maxdownfilterlength];
  
  
  Size=OutputBufferSize+filterlength/DownsampleAmnt+1;

  if (FCB->WorkBufSize != Size) {
    if (FCB->WorkBuf) free(FCB->WorkBuf);
    FCB->WorkBufSize=Size;
    /* allocate and fill with zeros */
    FCB->WorkBuf=calloc(FCB->WorkBufSize*FCB->NumChannels*sizeof(INT32),1);
  }

  Ptr=FCB->WorkBuf;

  /* Do the filtering */

  for (Cnt=0;Cnt < InputBufferSize;Cnt++) {
    for (Channel=0;Channel < NumChannels; Channel++) {
      Value=InputBuf[Cnt*NumChannels+Channel];
      for (Cnt2=((Cnt%DownsampleAmnt)==0) ? 0:(DownsampleAmnt-(Cnt%DownsampleAmnt));Cnt2 < filterlength;Cnt2+=DownsampleAmnt) {
	/*if (((Cnt+Cnt2) % DownsampleAmnt) != 0) fprintf(stderr,"foo!\n"); */
	Ptr[(Cnt+Cnt2)/DownsampleAmnt*NumChannels+Channel]+=Value*filter[Cnt2];
      }
    }
  }
  
  /* done the filtering. Now extract the filtered data into the outputbuf, and move the remains
     into the start of the WorkBuf, where it will be used by the next filtering operation */
  
  for (Cnt=0; Cnt < OutputBufferSize;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      OutputBuf[Cnt*NumChannels+Channel]=Ptr[Cnt*NumChannels+Channel] >> 15; /* only take high-order 17 bits */
    }
  }
  

  /* Move tail end: */
  
  for (Cnt=OutputBufferSize;Cnt < Size;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      Ptr[(Cnt-OutputBufferSize)*NumChannels+Channel]=Ptr[Cnt*NumChannels+Channel];
    }
  }
  
  /* clear everything else */
  for (Cnt=Size-OutputBufferSize;Cnt < Size;Cnt++) {
    for (Channel=0;Channel < NumChannels;Channel++) {
      Ptr[Cnt*NumChannels+Channel]=0;
    }
  }

  return FCB;
}

void ClearFCB(struct FilterControlBlock *FCB) /* zero out any stored response in the FCB */
{
  if (FCB->WorkBuf && FCB->WorkBufSize)
    memset(FCB->WorkBuf,0,FCB->WorkBufSize*FCB->NumChannels*sizeof(INT32));
  
}


