/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* this module communicates with the Linux (and other Unix) 'Open Sound System'
   audio API. 

   See http://www.4front-tech.com for more information 

   */


#ifdef LINUX_OSS  /* only compile this module if OSS drivers selected */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <string.h>
#include <errno.h>

#include "io_spec_old.h"
#include "soundspec.h"
#include "linklist.h"
#include "util.h"
#include "soundcardio.h"

extern char **gargv;


struct soundspec *sound_init(int CardID,int CardDevice,int Mode,int log2fragsize,int log2numfrags)
     /* returns sound file descriptor */
     /* this must be followed by a call to sound_setrate() once rate information is available. */
     /* if log2fragsize or log2numfrags is less than zero, the sound driver
	internal default fragment size is used. */
     /* CardDevice is not used (it is for linux_alsa.c compatibility) */
{
  int OpenMode=0;
  int msg,format;
  char *testptr;
  int audiofd;
  char fname[]="/dev/dspi";
  struct soundspec *spec;
  struct audio_buf_info info;

  spec=malloc(sizeof(struct soundspec));
  spec->CardID=CardID;
  spec->Mode=Mode;
  spec->log2fragsize=log2fragsize;
  spec->log2numfrags=log2numfrags;
  spec->Rate=0;
  spec->NumChannels=0;

  if (Mode==SIM_RECORD) OpenMode=O_RDONLY;
  if (Mode==SIM_PLAY) OpenMode=O_WRONLY;
  if (Mode==SIM_DUPLEX) OpenMode=O_RDWR;

  if (CardID==0) fname[strlen(fname)-1]=0;
  else fname[strlen(fname)-1]='0'+CardID;
  /*fprintf(stderr,"opening %s\n",fname);*/
  audiofd=open(fname,OpenMode);
  /*  fprintf(stderr,"OpenMode=%ld\n",(long)OpenMode);*/

  /* set driver fragment (buffer) characteristics */
  if (log2numfrags >=0  && log2fragsize >=0) {
    msg=(log2numfrags << 16) | (log2fragsize); 
    
    /*fprintf(stderr,"fragsize=%ld\n",(long)(1<<log2fragsize)); */
    
    ioctl(audiofd, SNDCTL_DSP_SETFRAGMENT, &msg);
  }
  

  if (Mode==SIM_DUPLEX) {
    ioctl(audiofd,SNDCTL_DSP_SETDUPLEX, 0);
    msg=0;
    ioctl(audiofd, SNDCTL_DSP_GETCAPS, &msg);
    if (!(msg & DSP_CAP_DUPLEX)) {
      fprintf(stderr,"%s: Hardware does not support duplex mode.\n",gargv[0]);
      exit(1);
    }
    
  }
  /* set sample format  (16 bit, unsigned hardware-endian) */

  
  format=1;
  testptr=(char *) &format;
  if (*testptr) format=AFMT_S16_LE; /* little endian */
  else format=AFMT_S16_BE;
  msg=format;

  ioctl(audiofd, SNDCTL_DSP_SETFMT, &msg);
  if (msg != format) {
    fprintf(stderr,"Unable to set format to %ld. Got %ld\n",(long)format,(long)msg);
  }
  
  if (Mode==SIM_DUPLEX || Mode==SIM_PLAY) {
    ioctl(audiofd,SNDCTL_DSP_GETOSPACE,&info);    
    fprintf(stderr,"Input info: %d fragments, %d fragstotal, %d fragsize, %d bytes\n",(int)info.fragments,(int)info.fragstotal,(int)info.fragsize,(int)info.bytes);
  }
  if (Mode==SIM_DUPLEX || Mode==SIM_RECORD) {
    ioctl(audiofd,SNDCTL_DSP_GETISPACE,&info);    
    fprintf(stderr,"Output info: %d fragments, %d fragstotal, %d fragsize, %d bytes\n",(int)info.fragments,(int)info.fragstotal,(int)info.fragsize,(int)info.bytes);
  }
  spec->recordsoundfh=audiofd;
  spec->playsoundfh=audiofd; /* with OSS record & play always share the FD */
  
  setnonblocking(audiofd);

  return spec;


}

void sound_reopen(struct soundspec *spec)
{
  struct soundspec *spec1;
  int Num;
  int fargs;

  fargs=fcntl(spec->recordsoundfh,F_GETFL);
  
  close(spec->recordsoundfh);
  

  spec1=sound_init(spec->CardID,0,spec->Mode,spec->log2fragsize,spec->log2numfrags);
  if (fargs & O_NONBLOCK) {
    fprintf(stderr,"Setting new file descriptor to non-blocking mode\n");
    setnonblocking(spec1->recordsoundfh);
  }

  Num=spec->NumChannels;
  sound_setrecordrate(spec1,&Num,spec->Rate);
  if (Num != spec->NumChannels) {
    fprintf(stderr,"sound_reopen: error setting number of channels\n");
  }
  if (spec1->Rate != spec->Rate) {
    fprintf(stderr,"sound_reopen: error setting sampling rate. Desired=%d, actual=%d\n",(int)spec->Rate,(int)spec1->Rate);
  }
  *spec=*spec1;
  free(spec1);
}


int sound_setrecordrate(struct soundspec *spec,int *NumChannels,int DesiredRate)
     /* *numchannels=1 for mono, 2 for stereo */
     /* *numchannels will be set to 1 if the h/w doesn't support stereo */
     /* returns actual sample rate */
     /* Mode = SIM_RECORD, SIM_PLAY, or SIM_DUPLEX */
{
  int msg;
  int Speed;
  static int RecursiveFlag;

  fprintf(stderr,"got sound_setrecordrate() NumChannels=%d\n",(int)*NumChannels);
  if (DesiredRate==spec->Rate && *NumChannels==spec->NumChannels) return spec->Rate;


  /* this routine may be called recursively by itself if it gets an 
     error and tries to close and reopen the sound driver */

  if (RecursiveFlag > 1) {
    fprintf(stderr,"Error in reopening sound driver\n");
    exit(1);
  }
  RecursiveFlag++;

#define LINUX_OSS_DRIVER_BUG4_HACK
#ifdef LINUX_OSS_DRIVER_BUG4_HACK
  /* in BUG4_HACK we always close and reopen the soundcard rather than
     switching rates in midstream */
  
  if (spec->Rate && RecursiveFlag==1) {
    spec->Rate=DesiredRate;
    spec->NumChannels=*NumChannels;
    sound_reopen(spec);
    Speed=spec->Rate;
    *NumChannels=spec->NumChannels;
    RecursiveFlag=0;
    return Speed;
  }
#endif


  msg=0;
  ioctl(spec->recordsoundfh,SOUND_PCM_RESET,&msg); /* reset soundcard, so we can change parameters */
  /*  sleep(1); */
  
  /*  fprintf(stderr,"NumChannels=%ld\n",(long)*NumChannels);*/

  if (*NumChannels == 2) 
    msg=1;
  else
    msg=0;
  ioctl(spec->recordsoundfh,SNDCTL_DSP_STEREO,&msg);
  
  spec->NumChannels=*NumChannels;
  if (*NumChannels==2 && msg==0) {
    /* stereo not supported */
    spec->NumChannels=1;
    /**NumChannels=1;*/
  }

  if (*NumChannels == 2) fprintf(stderr,"Stereo!!!\n");
  
  
  /* set sample rate */
  msg=DesiredRate;
  fprintf(stderr,"Soundcard #%d: Desired rate=%d",(int)spec->CardID,(int)msg);
  ioctl(spec->recordsoundfh,SNDCTL_DSP_SPEED,&msg);
  Speed=msg;
  fprintf(stderr," Actual rate=%d\n",(int)msg);

  spec->Rate=Speed;
  

  /* because of a bug in the sound driver, we have to 
     initiate a read now, or it won't terminate the select() */
  {
    char *junkbuf;
    int sz;
    int nbytes;
    audio_buf_info inpinfo;
    audio_buf_info outinfo;
    int nread=0,nwrite=0;
    int nalloc;
    int nbytes1=0,nbytes2=0;

    /* read exactly one fragment */
    /*ioctl(spec->soundfh,SNDCTL_DSP_GETBLKSIZE,&nread);*/
    
    if (spec->Mode==SIM_RECORD || spec->Mode==SIM_DUPLEX) {
      ioctl(spec->recordsoundfh,SNDCTL_DSP_GETISPACE,&inpinfo);
      nread=inpinfo.fragsize;
    }
    if (spec->Mode==SIM_PLAY || spec->Mode==SIM_DUPLEX) {
      ioctl(spec->recordsoundfh,SNDCTL_DSP_GETOSPACE,&outinfo);
      nwrite=outinfo.fragsize*outinfo.fragstotal;
      
    }
    nalloc=maxva(nread,nwrite,-2);
    junkbuf=malloc(nalloc);
    

    if (spec->Mode == SIM_RECORD || spec->Mode == SIM_DUPLEX) {
      nbytes1=read(spec->recordsoundfh,junkbuf,nread);
      fprintf(stderr,"read %ld bytes\n",(long)nbytes1);
      
    }
#define LINUX_OSS_DRIVER_BUG3_HACK
#ifdef LINUX_OSS_DRIVER_BUG3_HACK
    if (spec->Mode==SIM_DUPLEX && nbytes1 > 0) {
      /* for some reason the driver doesn't work right 
	 unless we wait for at least one record block to be read
	 before going into play mode. If we wait too long it 
	 doesn't work either */
      fd_set writeset;
      FD_ZERO(&writeset);
      FD_SET(spec->recordsoundfh,&writeset);
      select(spec->recordsoundfh+1,&writeset,NULL,NULL,NULL);
      
    }
#endif
    /* Note that the above bug is made much worse if 
       you allow this conditional in duplex mode! */
    //    if ( spec->Mode==SIM_PLAY/* || spec->Mode==SIM_DUPLEX*/) {
    //  memset(junkbuf,0,nwrite);
    //  nbytes2=write(spec->soundfh,junkbuf,nwrite);
    //  fprintf(stderr,"wrote %ld bytes\n",(long)nbytes2);
    //}

    if (nbytes1 < 0 || nbytes2 < 0) {
      perror("sound_setrate_read_write");
      fprintf(stderr,"sound_setrate_read_write: trying to reopen soundcard...\n");
      spec->Rate=DesiredRate;
      spec->NumChannels=*NumChannels;
      sound_reopen(spec);
      Speed=spec->Rate;
    }
    free(junkbuf);
  }


  /*  fprintf(stderr,"Set record sample rate %ld\n",(long)Speed); */


  RecursiveFlag=0;

  *NumChannels=spec->NumChannels;
  return Speed; /* return actual sample rate. */
}

int sound_setplayrate(struct soundspec *spec,int *NumChannels,int DesiredRate)
     /* *numchannels=1 for mono, 2 for stereo */
     /* *numchannels will be set to 1 if the h/w doesn't support stereo */
     /* returns actual sample rate */
     /* Mode = SIM_RECORD, SIM_PLAY, or SIM_DUPLEX */
{
  int Msg;
  if (spec->Mode==SIM_PLAY) { /* play only -- we can let the playing do rate-determination */
    
    fprintf(stderr,"HW rate=%ld\n",(long)DesiredRate);
    return sound_setrecordrate(spec,NumChannels,DesiredRate);
  } else {
    /*ioctl(soundfh,SOUND_PCM_READ_RATE,&Msg);*/
    return spec->Rate;
  }
}



#if 0  /* this next version of sound_readbuf would work if the sounddriver
	     weren't braindead */

int sound_readbuf(int soundfh,INT16 *CurBlock,int Size,int CurPos,int *moreavail,int DebugFlag)
     /* returns new CurPos. Block is finished if CurPos==Size */
     /* size and CurPos are in bytes */
{
  int GotBytes;
  /* read in from soundcard */

  
  
  while(CurPos < Size) {
    
    GotBytes=read(soundfh,((char *)(CurBlock))+CurPos,Size-CurPos);
    fprintf(stderr,"sound_readbuf: got %ld bytes\n",(long)GotBytes);
    if (GotBytes < 0 && errno != EWOULDBLOCK) {
      fprintf(stderr,"%s: Read error on input channel: %d\n",gargv[0],(int)errno);
      exit(1);
    }
    else if (!GotBytes) {
      fprintf(stderr,"%s: EOF on input channel\n",gargv[0]);
      exit(1);
    }
    else if (GotBytes < 0 && (errno==EWOULDBLOCK || errno==EINTR || errno==EAGAIN)) {
      break; /* go back to select() */
    }
    CurPos+=GotBytes;

  }
 
  
  return CurPos;
}

#endif

int sound_readbuf(struct soundspec *spec,INT16 *CurBlock,int Size,int CurPos,int *moreavail,int DebugFlag)
     /* returns new CurPos. Block is finished if CurPos==Size */
     /* size and CurPos are in bytes */
{
  int AskBytes,GotBytes;
  audio_buf_info info;
  float debugtime;


  /* read data from soundcard */

  
  ioctl(spec->recordsoundfh,SNDCTL_DSP_GETISPACE,&info);
  
  /* info.fragsize is size of a fragment. */
  /* info.fragstotal is total number of fragments in buffer */
  /* info.fragments is number of fragments available */
  /* info.bytes is number of bytes that can be read or written 
     immediately without blocking */

  if (DebugFlag) fprintf(stderr,"record: info.fragsize=%ld info.bytes=%ld info.fragments=%ld availspace=%ld\n",(long)info.fragsize,(long)info.bytes,(long)info.fragments,(long)(Size-CurPos));

  
  AskBytes=info.bytes;
  /*AskBytes=info.fragsize*info.fragments;*/

  if (AskBytes > (Size-CurPos)) {
    AskBytes=Size-CurPos;
    *moreavail=1;
  } else {
    *moreavail=0;
  }

  startdebugtimer();
  
  if (AskBytes > 0) {
    do {
      GotBytes=read(spec->recordsoundfh,((char *)CurBlock)+CurPos,AskBytes);
    } while (GotBytes < 0 && (errno==EINTR || errno==EAGAIN));
    debugtime=resultdebugtimer();
    if (debugtime > 5e-3) fprintf(stderr,"Slow read: %f ms. AskBytes=%ld \n",(float)(debugtime*1000),(long)AskBytes);

    /*   fprintf(stderr,"Got %ld bytes (of %ld) from soundcard in %g ms\n",(long)GotBytes,(long)(Size-CurPos),(float)(resultdebugtimer()*1000));
     */
    
    if (GotBytes <= 0 && errno != EWOULDBLOCK) {
      fprintf(stderr,"%s: Write error on output channel: %d\n",gargv[0],(int)errno);
      exit(1);
    }
    
    if (GotBytes > 0)
      CurPos+=GotBytes;
  }
  
  
  return CurPos;
}



#if 0  /* this next version of sound_readbuf would work if the sounddriver
	     weren't braindead */

int sound_writebuf(int soundfh,INT16 *CurBlock,int Size,int CurPos)
{
  int SentBytes;
  /* write data to soundcard */
 
  /* fprintf(stderr,"Got sound_writebuf()\n");*/
  while (CurPos < Size) {
    startdebugtimer();
    SentBytes=write(soundfh,((char *)CurBlock)+CurPos,Size-CurPos);
    fprintf(stderr,"Wrote %ld bytes to soundcard in %g ms\n",(long)SentBytes,(float)(resultdebugtimer()*1000));
    if (SentBytes <= 0 && errno != EWOULDBLOCK) {
      fprintf(stderr,"%s: Write error on output channel: %d\n",gargv[0],(int)errno);
      exit(1);
    }
    else if (SentBytes <= 0 && (errno==EWOULDBLOCK || errno==EINTR || errno==EAGAIN)) {
      break; /* go back to select() */
    }
    CurPos+=SentBytes;

  }
  
  
  return CurPos;
  
}

#endif


int sound_writebuf(struct soundspec *spec,INT16 *CurBlock,int Size,int CurPos,int *morespace,int DebugFlag)
     /* puts 1 in *morespace if the soundcard buffer has more space available
	0 otherwise */
{
  int SendBytes,SentBytes;
  audio_buf_info info;
  float debugtime;
  static int TotalData;
  
  /* write data to soundcard */
  *morespace=0;
  
  ioctl(spec->recordsoundfh,SNDCTL_DSP_GETOSPACE,&info);
  
  /* info.fragsize is size of a fragment. */
  /* info.fragstotal is total number of fragments in buffer */
  /* info.fragments is number of fragments available */
  /* info.bytes is number of bytes that can be read or written 
     immediately without blocking */

  if (DebugFlag) fprintf(stderr,"play: info.fragsize=%ld info.bytes=%ld info.fragments=%ld availdata=%ld\n",(long)info.fragsize,(long)info.bytes,(long)info.fragments,(long)(Size-CurPos));

  SendBytes=info.bytes;
  /*SendBytes=info.fragsize*info.fragments;*/

  if (SendBytes > (Size-CurPos)) {
    *morespace=1;
    SendBytes=Size-CurPos;
  }

  startdebugtimer();
  
  if (SendBytes > 0) {
    do {
      SentBytes=write(spec->recordsoundfh,((char *)CurBlock)+CurPos,SendBytes);
    } while (SentBytes < 0 && (errno==EINTR || errno==EAGAIN));
    debugtime=resultdebugtimer();
    if (debugtime > 5e-3) fprintf(stderr,"Slow write\n");
    
    /*fprintf(stderr,"Wrote %ld bytes (of %ld) to soundcard in %g ms\n",(long)SentBytes,(long)(Size-CurPos),(float)(resultdebugtimer()*1000));*/
    
    
    if (SentBytes <= 0 && errno != EWOULDBLOCK) {
      fprintf(stderr,"%s: Write error on output channel: %d\n",gargv[0],(int)errno);
      exit(1);
    }
    
    if (SentBytes > 0)
      CurPos+=SentBytes;
    TotalData+=SentBytes;
    /*    fprintf(stderr,"TotalData=%ld\n",(long)TotalData);*/
  }
  
  return CurPos;
  
}

#endif /* LINUX_OSS */
