/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "io_spec_old.h"
#include "soundspec.h"
#include "linklist.h"
#include "util.h"
#include "filter.h"
#include "soundcardio.h"
#include "panelinterface_h.h"


/* known bug (FIXED): when recording, at the beginning it outputs two adjacent header
   blocks rather than one. Need to do a protocol check on the output */


/* Listing of buffers
   ------------------

   RecordBuffer   --  Data read in from soundcard
   RecordTransmitBuffer -- Data ready to be transmitted to record output pipe,
                           including header and status blocks
   PlayReceiveBuffer -- Data received from the play input pipe, including
                        header and status blocks
   PlayBuffers -- a series of buffers used to pad between the input and
                  output
   PlayOutputBuffer -- the buffer used to output directly to the soundcard. 
                       data in this buffer may occasionally have extra samples
                       inserted or samples removed so as to keep the 
		       PlayBuffers buffer-set half full. This is so that
		       the system will work properly even when there are small
		       differences between the record and play sample rates. 
   MonitorOutputBuf -- This buffer collects data from immediately after 
                       upsampling of data to be played. It is sent directly
		       to the monitor port (#3). It also contains header
		       and status blocks. 

   */
  

/* #define DEBUG_BUFFERING */

#ifdef LINUX_OSS
#define LINUX_OSS_DRIVER_BUG_HACK
#define LINUX_OSS_DRIVER_BUG2_HACK
#endif


int RecordMaxRate=44100;


int RecordDebug=0;
int PlayDebug=0;


int PlayMaxRate=44100;

int selecttimeout=50000; /* 50000 = 50ms */

int PlayPipe=0;
int RecordPipe=1;
int MonitorPipe=3;

/*int playsoundcardfd=-1;
int recordsoundcardfd=-1;
*/
struct soundspec *playsoundcard=NULL;
struct soundspec *recordsoundcard=NULL;


int NoMonitor=0;
int DrainSync=0; /* if drainsync is enabled, the player will operate as a 
		    synchronization source */


int RecordCardNum=0;
int RecordCardDev=0;
int PlayCardNum=0;
int PlayCardDev=0;

int Mode=SIM_DUPLEX;

int PlayMode=SIM_DUPLEX; /* Mode of soundcard which is playing */
int RecordMode=SIM_DUPLEX; /* Mode of soundcard which is recording */
/*
int recordlog2fragsize=10;
int recordlog2numfrags=3;
*/

int recordlog2fragsize=-1; /* -1 means use driver/soundcard internal defaults */
int recordlog2numfrags=-1;

int playlog2fragsize=-1; /* -1 means use driver/soundcard internal defaults */
int playlog2numfrags=-1;

/* These three are specified */

int RecordSampleRate=44100; /* sample rate the user has requested (after downsampling) */

#define DAC_IMPULSE 0
#define DAC_STEP 1
#define DAC_LINEAR 2
#define DAC_FILTERED 3
int DACMode=DAC_FILTERED;
int RecordNumChannels=1; /* 2 for stereo, 1 for mono */


int RecordHWSampleRate=44100;
int RecordDownSample=1;


#define ADC_UNFILTERED 0
#define ADC_FILTERED 1


int PlayConfigured=0; /* Set to 1 once play sample rate has been set */
/* These three are specified */
int ADCMode=ADC_UNFILTERED;
int DesiredPlaySampleRate=44100; /* Sample rate requested by the input stream */
int PlayNumChannels=2; /* 2 for stereo, 1 for mono */

/* And these two are calculated from that. Note that HWRate != SampleRate*UpSample in general because the soundcard may
   be incapable of playing at precisely the desired rate */
int PlaySampleRate=44100; /* actual sample rate obtained. Should be close to DesiredPlaySampleRate */
int PlayHWSampleRate=44100;
int PlayUpSample=1;



INT16 *RecordBuffer; /* stores data just collected from the soundcard */
int RecordBufferSize=1024; /* number of 16 bit (stereo) samples in buffer */
int RecordBufferBytes; /* = RecordBufferSize*2*numchannels */
int CurRecordPos; /* in bytes */

INT16 *RecordTransmitBuffer; /* stores data ready to be sent down the output pipe */

int RTLength; /* in 16 bit (stereo) samples. Allocated size is actually RTLength*2*NumChannels+sizeof(struct DataBlkHeader) */
int RTBytes; /* size of data currently going out, in bytes */
int RTPos; /* in bytes */

int RecordStatusPromised=1;

struct FilterControlBlock *RecordBufFCB=NULL;

INT16 RecordFrameNumber=1;
/*INT16 RecordStatusPeriod=100;*/ /* Send a status block once every this-many blocks */
INT16 RecordStatusPeriod=1; /* Sending a status block every time makes us
			       compatible with the new spec */


INT16 *PlayBuffers; /* stores data about to be sent to the soundcard */
int PlayBufferSize=1024; /* number of 16 bit (stereo) samples in buffer */
int PlayBufferBytes; /* PlayBufferSize*2*NumChannels*/
int NumPlayBuffers=15;
struct FilterControlBlock *PlayBufFCB=NULL;

int FPBTarget=-1; /* Number of free play buffers we are waiting for before we start playing
		     (-1 = ok to play) */


#define PlayBuffer(num) (PlayBuffers+(num)*PlayBufferSize*PlayNumChannels)
#define PlayBufferElem(num,index) (*(PlayBuffers+(index)+(num)*PlayBufferSize*PlayNumChannels))

int NextAvailBuffer; /* Which buffer to fill up next */

int NextPlayBuffer; /* Next buffer to copy to PlayOutputBuffer */


int SyncGain=1; /* gain factor for use in synchronization see CopyPBtoPOB() */
int InsertPointThreshold=10; /* How close to end of block can we insert a point */
int RemovePointThreshold=4; /* ... remove a point. See CopyPBtoPOB() */

INT16 *PlayOutputBuffer; /* Stores data about to be written to the 
			     soundcard. */
int POBSize; /* Number of samples for which there is space in the 
		PlayOutputBuffer. Should be greater than PlayBufferSize 
		(typically PlayBufferSize*3/2) to provide space for 
		interpolated points that have been added for synchronization
		reasons. */
int POBBytes; /* number of bytes in the block currently going out. Must be 
		 <= POBSize*2*NumChannels */
int POBPos; /* Current position in the block currently going out. */

INT16 *PlayReceiveBuf; /* Stores data just collected from the pipe, to be played */

int PRBSize; /* number of 16 bit (stereo) samples in buffer. Allocated size is actually PRBSize*2*NumChannels+sizeof(struct DataBlkHeader) */
int PRBBytes; /* number of bytes in this block */
int PRBNextBytes; /* number of bytes in the next block. Either PRBSize*2*NumChannels+datablkheadersize or sizeof(struct StatusBlk) */
int PRBPos; /* position in the buffer, in bytes */ 


INT16 *MonitorOutputBuf; /* Data sent   setnonblocking(readfromgui);
			    here after upsampling, for post-upsample 
			     monitoring */
int MOBSize; /* number of 16 bit (stereo) samples for which buffer
		was allocated. Allocated size is actually MOBSize*2*NumChannels + sizeof(DataBlkHeader) */
int MOBBytes; /* Number of bytes used in the block currently being transmitted */
int MOBPos;
int MOBPromisedStatusBlk=1; /* Set to 1 if we have promised a status block
			       as the next block to be output on MOB.
			       If we didn't promise it, we can't send it! */

int MonitorOutput=0; /* MonitorOutput is 1 if we are doing the monitor output thing */
int MonitorOutputFID=3; /* pipe number for monitor output */

char *Title="SoundcardIO"; /* Desired window title for control panel */
char **gargv;


int panelinterfacereadfd=-1;
struct queued_outbuffer *panelinterfacewrite=NULL;
int NoPanel=0;

/* For queueing changes to record parameters: */
struct QueuedRecordParams {
  int Valid; /* are the parameters actually queued? */
  int NewBufSize;
  int NewNumChannels;
  int NewRecordRate;
} QRP={0,0,0,0};

/* For queueing changes to play parameters: */
struct QueuedPlayParams {
  int Valid; /* are the parameters actually queued? */
  int NewNumBuffers;
} QPP={0,0};

void BuildRecordBuffers(int NewSize)
     /* times are desired record buffering times in ms */
{
  /* Buffer size should be a multiple of 4, a multiple of the DownSampleAmount, 
     and correspond to about BufferTime of delay */

  /* preferably also slightly less than a power of 2 in size, like 220 or so */
  /* but we don't calculate that automatically. */
  

  /*NewSize=((BufferTime*RecordSampleRate/1000)/DownSampleAmount/4)*DownSampleAmount*4;*/

  /* Temporary for debugging: */
  /*NewSize=8192/2/RecordNumChannels;*/
  
  if (RecordBuffer) free(RecordBuffer);
  RecordBufferSize=(NewSize/RecordDownSample)*RecordDownSample;
  RecordBufferBytes=RecordBufferSize*2*RecordNumChannels;
  RecordBuffer=malloc(RecordBufferSize*2*RecordNumChannels);
  CurRecordPos=0;
  
  RTLength=NewSize/RecordDownSample;
  RecordTransmitBuffer=malloc(RTLength*2*RecordNumChannels+sizeof(struct DataBlkHeader));
  RTPos=0;
  /*  fprintf(stderr,"RTLength=%d\n",(int)RTLength);*/
}

void BuildPlayBuffers(int BlockSize,int NumBuffers)
{
  
  int NewSize;
  int Cnt,Cnt2;

  /* NewSize=((BufferTime*PlaySampleRate/1000)/UpSampleAmount/4)*UpSampleAmount*4;*/
  NewSize=BlockSize*PlayUpSample;
  
  if (PlayBuffers) free(PlayBuffers);
  PlayBufferSize=NewSize;
  PlayBufferBytes=PlayBufferSize*2*PlayNumChannels;
  PlayBuffers=malloc(PlayBufferSize*2*PlayNumChannels*NumBuffers);
  NumPlayBuffers=NumBuffers;

  NextAvailBuffer=0;
  NextPlayBuffer=0;
  
  PRBSize=NewSize/PlayUpSample;
  PlayReceiveBuf=malloc(PRBSize*2*PlayNumChannels+sizeof(struct DataBlkHeader));
  PRBPos=0;
  PRBBytes=0;
  
  POBSize=NewSize*3/2;
  POBBytes=0;
  POBPos=0;
  PlayOutputBuffer=malloc(POBSize*2*PlayNumChannels+sizeof(struct DataBlkHeader));

  MOBSize=NewSize;
  MOBBytes=0;
  MOBPos=0;
  MonitorOutputBuf=malloc(MOBSize*2*PlayNumChannels+sizeof(struct DataBlkHeader));

  if (panelinterfacewrite) {
    qob_printf(panelinterfacewrite,1,"PlayBufferSize=%d\n",(int)(PlayBufferSize*NumPlayBuffers));
  }

  /* Let the newly allocated buffer fill up halfway before starting playback */
  FPBTarget=NumPlayBuffers/2;

#ifdef DEBUG_BUFFERING
  fprintf(stderr,"BuildPlayBuffers: PlayBufferSize=%d, NumPlayBuffers=%d\n",(int)PlayBufferSize,(int)NumPlayBuffers);
#endif
}

void CalcRecordRates(int NewRecordSampleRate,int NewNumChannels)
{
  
  RecordSampleRate=NewRecordSampleRate;
  /* RecordDownSample=(RecordMaxRate/RecordSampleRate);*/

  for (RecordDownSample=1;RecordSampleRate*RecordDownSample <= RecordMaxRate/2;RecordDownSample<<=1);
  
  RecordHWSampleRate=RecordSampleRate*RecordDownSample;
  
  RecordNumChannels=NewNumChannels;
  RecordHWSampleRate=sound_setrecordrate(recordsoundcard,&RecordNumChannels,RecordHWSampleRate);
  RecordSampleRate=RecordHWSampleRate/RecordDownSample;

  if (panelinterfacewrite) {
    qob_printf(panelinterfacewrite,1,"RecordSampleRate=%d\n",(int)RecordSampleRate);
    qob_printf(panelinterfacewrite,1,"RecordNumChannels=%d\n",(int)RecordNumChannels);
    
  }
}

void InitQRP(void)
{
  if (!QRP.Valid) {
    QRP.Valid=1;
    QRP.NewBufSize=RecordBufferSize;
    QRP.NewNumChannels=RecordNumChannels;
    QRP.NewRecordRate=RecordSampleRate;
  }
}

void QueueNewRecordRate(int NewSampleRate)
{
  InitQRP();
  QRP.NewRecordRate=NewSampleRate;
}

void QueueNewRecordNumChannels(int NewNumChannels)
{
  InitQRP();
  QRP.NewNumChannels=NewNumChannels;
  
}

void QueueNewRecordBufSize(int NewBufferSize)
{
  InitQRP();
  QRP.NewBufSize=NewBufferSize;  
}

void InitQPP(void)
{
  QPP.Valid=1;
  QPP.NewNumBuffers=NumPlayBuffers;
}

void QueueNewPlayNumBuffers(int newnum)
{
  InitQPP();
  QPP.NewNumBuffers=newnum;
}

void SetADCMode(int Newmode)
{
  if (Newmode != ADCMode) {
    ADCMode=Newmode;
    if (RecordBufFCB) ClearFCB(RecordBufFCB);
    if (panelinterfacewrite) qob_printf(panelinterfacewrite,1,"ADCMode=%d\n",(int)ADCMode);
  }
}

void SetDACMode(int Newmode)
{
  if (Newmode != DACMode) {
    DACMode=Newmode;
    if (PlayBufFCB) ClearFCB(PlayBufFCB);
    if (panelinterfacewrite) qob_printf(panelinterfacewrite,1,"DACMode=%d\n",(int)DACMode);
  }
  
}

void CalcPlayRates(int NewPlaySampleRate,int NumChannels)
{
  DesiredPlaySampleRate=NewPlaySampleRate;
  PlaySampleRate=NewPlaySampleRate;
  PlayNumChannels=NumChannels;
  /*PlayUpSample=(PlayMaxRate/PlaySampleRate);*/
  for (PlayUpSample=1;PlaySampleRate*PlayUpSample <= PlayMaxRate/2;PlayUpSample<<= 1);

#ifdef DEBUG_BUFFERING
  fprintf(stderr,"PlayUpSample=%d\n",(int)PlayUpSample);
#endif

  PlayHWSampleRate=PlaySampleRate*PlayUpSample;
  PlayHWSampleRate=sound_setplayrate(playsoundcard,&PlayNumChannels,PlayHWSampleRate);
  PlaySampleRate=PlayHWSampleRate/PlayUpSample;
  /*fprintf(stderr,"PlayHWSampleRate=%d\n",(int)PlayHWSampleRate);*/
  if (panelinterfacewrite) {
    qob_printf(panelinterfacewrite,1,"PlaySampleRate=%d\n",PlaySampleRate);
    qob_printf(panelinterfacewrite,1,"PlayNumChannels=%d\n",PlayNumChannels);
  }
}


int WriteRTBuf(void)
     /* Write the record transmit buffer to its output pipe */
     /* returns one if the buffer is empty after the call, zero otherwise */
{
  int NumBytes;

  while (RTPos < RTBytes) {
    NumBytes=write(RecordPipe, ((char *)RecordTransmitBuffer)+RTPos,RTBytes-RTPos);
    if (NumBytes < 0 && (errno==EWOULDBLOCK || errno==EAGAIN || errno==EINTR)) {
      /* can't write -- return !! */
      return 0;
    } else if (NumBytes < 0) {
      /* Error !!! */
      perror("SoundcardIO/WriteRTBuf");
      exit(1);
    }
    RTPos+=NumBytes;
  }
  
  return 1;

}


void WriteRecordStatus(void)
     /* Note: The RecordTransmitBuffer should be flushed before 
	calling this function.
	If you actually want to write it out, you should call WriteRTBuf() */
{
  struct StatusBlk Status={ { 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f },
			    0,
			    0,
			    0,
			    0,
			    0};
  Status.NumChannels=RecordNumChannels;
  Status.SampleRate=RecordSampleRate;
  Status.FrameNumber=RecordFrameNumber;
  Status.BlockSize=RTLength;

  memcpy(RecordTransmitBuffer,&Status,sizeof(struct StatusBlk));
  RTBytes=sizeof(struct StatusBlk);
  RTPos=0;
}


void ReadPRBuf(void)
     /* read any data available on the input pipe into the PlayReceiveBuffer */
{
  int GotBytes;
  
  /*  fprintf(stderr,"ReadPRBuf: PRBPos=%d, PRBBytes=%d...",PRBPos,PRBBytes);*/
  while(PRBPos < PRBBytes) {
    GotBytes=read(PlayPipe,((char *)(PlayReceiveBuf))+PRBPos,PRBBytes-PRBPos);
    /*    fprintf(stderr,"Read %ld bytes from input pipe, of %ld, PRBBytes=%ld\n",(long)GotBytes,(long)(PRBBytes-PRBPos),(long)PRBBytes);*/
    if (GotBytes < 0 && errno != EWOULDBLOCK) {
      fprintf(stderr,"%s: Read error on input channel: %d\n",gargv[0],(int)errno);
      exit(1);
    }
    else if (!GotBytes) {
      fprintf(stderr,"%s: EOF on input channel\n",gargv[0]);
      exit(1);
    }
    else if (GotBytes < 0 && (errno==EWOULDBLOCK || errno==EINTR || errno==EAGAIN)) {
      break; /* go back to select() */
    }
    PRBPos+=GotBytes;

  }

  /*  fprintf(stderr,"PRBPos=%d\n",PRBPos);*/

}

void WriteMOBuf(void)
     /* Write any data in the MonitorOutputBuffer onto the monitor pipe */
{
  int NumBytes;

  while (MOBPos < MOBBytes) {
    NumBytes=write(MonitorPipe, ((char *)MonitorOutputBuf)+MOBPos,MOBBytes-MOBPos);
    if (NumBytes < 0 && (errno==EWOULDBLOCK || errno==EAGAIN || errno==EINTR)) {
      /* can't write -- return !! */
      return;
    } else if (NumBytes < 0) {
      /* Error !!! */
      perror("SoundcardIO/WriteMOBuf");
      exit(1);
    }
    MOBPos+=NumBytes;
  }  
}
void UpsamplePlayBuffer(void)
     /* Copy the (full) play receive buffer to the (empty) Monitor Output
	Buffer (if applicable) and to the next available PlayBuffer. 
	If no PlayBuffers are available the data is discarded. 
	Does not reset the PRB */
     /* Assume the MOB is empty */
{
  int Cnt,Cnt2;
  INT16 S0,S1;
  int Channel;

  if ((NextAvailBuffer + 1) % NumPlayBuffers != NextPlayBuffer) { /* only copy data if we won't overflow */
    
    /*    if (PlayDebug) fprintf(stderr,"PB #%d filled\n",NextAvailBuffer);*/

    if (DACMode==DAC_FILTERED || DACMode==DAC_LINEAR) {
      /*write(2,PlayReceiveBuf+sizeof(struct DataBlkHeader)/sizeof(INT16),PlayBufferSize*4);*/
      PlayBufFCB=FilterAndUpsample(PlayReceiveBuf+sizeof(struct DataBlkHeader)/sizeof(INT16),PlayBuffer(NextAvailBuffer),PlayUpSample,PlayNumChannels,PlayBufferSize,PlayBufFCB,(DACMode==DAC_LINEAR));
      /*      write(2,PlayBuffer(NextAvailBuffer),PlayBufferSize*4);*/
    }
    else if (DACMode==DAC_IMPULSE || DACMode==DAC_STEP/* || DACMode==DAC_LINEAR*/) {
      assert(PRBBytes==PRBSize*2*PlayNumChannels+sizeof(struct DataBlkHeader));
      assert(PlayBufferSize==PRBSize*PlayUpSample);
      
      /* First copy & upsample to the PlayBuffer */
      for (Cnt=0;Cnt < PRBSize;Cnt++) {
	for (Channel=0;Channel < PlayNumChannels;Channel++) {
	  PlayBufferElem(NextAvailBuffer,Cnt*PlayUpSample*PlayNumChannels+Channel)=PlayReceiveBuf[Cnt*PlayNumChannels+Channel+sizeof(struct DataBlkHeader)/sizeof(INT16)];
	  switch (DACMode) {
	  case DAC_IMPULSE:
	    for (Cnt2=1;Cnt2 < PlayUpSample;Cnt2++) {
	      PlayBufferElem(NextAvailBuffer,Cnt*PlayUpSample*PlayNumChannels+Cnt2*PlayNumChannels+Channel)=0; /* in signed mode, 0 is the zero level */
	    }
	    break;
	    
	  case DAC_STEP:
	    for (Cnt2=1;Cnt2 < PlayUpSample;Cnt2++) {
	      PlayBufferElem(NextAvailBuffer,Cnt*PlayUpSample*PlayNumChannels+Cnt2*PlayNumChannels+Channel)=PlayReceiveBuf[Cnt*PlayNumChannels+Channel+sizeof(struct DataBlkHeader)/sizeof(INT16)];
	    }
	    break;
	    
	    /*case DAC_LINEAR: *//* Linear interpolation (except at the block ends) */
	    /*    if (Cnt==PRBSize-1) {*/ /* step mode for last element in the block */
	    /* for (Cnt2=1;Cnt2 < PlayUpSample;Cnt2++) {
		PlayBufferElem(NextAvailBuffer,Cnt*PlayUpSample*PlayNumChannels+Channel+Cnt2*PlayNumChannels)=PlayReceiveBuf[Cnt*PlayNumChannels+Channel+sizeof(struct DataBlkHeader)/sizeof(INT16)];
	      }
	    } else {
	      S0=PlayReceiveBuf[Cnt*PlayNumChannels+Channel+sizeof(struct DataBlkHeader)/sizeof(INT16)];
	      S1=PlayReceiveBuf[(Cnt+1)*PlayNumChannels+Channel+sizeof(struct DataBlkHeader)/sizeof(INT16)];
	      
	    */ /* Do linear interp. */
	    /*	      for (Cnt2=1;Cnt2 < PlayUpSample;Cnt2++) {
		PlayBufferElem(NextAvailBuffer,Cnt*PlayNumChannels*PlayUpSample+Cnt2*PlayNumChannels+Channel)=S0 + (S1-S0)*Cnt2/PlayUpSample;
	      }
	    }
	    
	    break;
	    */
	  }
	}
	
	
      }
    }

    if (!NoMonitor) {
      assert(MOBSize==PlayBufferSize);
      
      /* First copy header straight */
      for (Cnt=0;Cnt < sizeof(struct DataBlkHeader)/sizeof(INT16);Cnt++) {
	MonitorOutputBuf[Cnt]=PlayReceiveBuf[Cnt];
      }
      if (MOBPromisedStatusBlk) {
	fprintf(stderr,"soundcardio: MOB protocol generation error\n");
      }
      MOBPromisedStatusBlk=((struct DataBlkHeader *)MonitorOutputBuf)->NextBlockIsStatusBlk;
      
      /*      fprintf(stderr,"MOBPromised=%d\n",MOBPromisedStatusBlk);*/

      /* now copy actual data from upsampled copy built above */
      for (Cnt=0;Cnt < MOBSize*PlayNumChannels;Cnt++) {
	MonitorOutputBuf[Cnt + sizeof(struct DataBlkHeader)/sizeof(INT16)]=PlayBufferElem(NextAvailBuffer,Cnt);
      }
      
      MOBBytes=MOBSize*2*PlayNumChannels+sizeof(struct DataBlkHeader);
      MOBPos=0;
    }
    
    /* Signal that we have put data into the next play buffer */
    NextAvailBuffer=(NextAvailBuffer+1)%NumPlayBuffers;      
    
    
    
  } else {
#ifdef DEBUG_BUFFERING
    fprintf(stderr,"Data lost due to synchronization problem between input and output\n");
#endif
    /* Now we lose even more data, in order to get back into synchronization */
    NextPlayBuffer+=NumPlayBuffers/2;
    NextPlayBuffer=(NextPlayBuffer)%NumPlayBuffers;
  }
}

void DownsampleRecordBuffer(void)
     /* Copy the (full) record buffer to the (empty) RecordTransmitBuffer,
	downsampling and filtering if necessary. 
	This routine also fills in the DataBlkHeader, and sets the 
	recordbuffer to empty and ready for another block. */
{
  int Cnt;
  struct DataBlkHeader *Hdr;
  
  assert(RecordBufferSize==RTLength*RecordDownSample);
  assert(RecordBufferBytes==RecordBufferSize*2*RecordNumChannels);
  
  /* First write header */
  Hdr=(struct DataBlkHeader *)RecordTransmitBuffer;
  Hdr->MagicByte[0]=Hdr->MagicByte[1]=Hdr->MagicByte[2]=0x7f;
  if ((((RecordFrameNumber) % RecordStatusPeriod)==0) || QRP.Valid) {
    Hdr->NextBlockIsStatusBlk=1;
    RecordStatusPromised=1;
  }
  else {
    Hdr->NextBlockIsStatusBlk=0;
    RecordStatusPromised=0;
  }
  

  if (ADCMode==ADC_UNFILTERED) {
    /*fprintf(stderr,"RecordDownSample=%ld\n",(long)RecordDownSample);*/
    for (Cnt=0;Cnt < RTLength;Cnt++) {
      RecordTransmitBuffer[Cnt*RecordNumChannels+sizeof(struct DataBlkHeader)/sizeof(INT16)]=RecordBuffer[Cnt*RecordNumChannels*RecordDownSample];
      if (RecordNumChannels == 2) RecordTransmitBuffer[Cnt*RecordNumChannels+1+sizeof(struct DataBlkHeader)/sizeof(INT16)]=RecordBuffer[1+Cnt*RecordNumChannels*RecordDownSample];
    }
    
  } else if (ADCMode==ADC_FILTERED) {
    /*fprintf(stderr,"Performing filtering\n");*/
    RecordBufFCB=FilterAndDownsample(RecordBuffer,RecordTransmitBuffer+sizeof(struct DataBlkHeader)/sizeof(INT16),RecordDownSample,RecordNumChannels,RecordBufferSize,RecordBufFCB);
  }

  CurRecordPos=0;
  RTBytes=RTLength*2*RecordNumChannels+sizeof(struct DataBlkHeader);
  RTPos=0;
  
}

int CopyPBtoPOB(int Bufnum)
     /* Copy the specified PlayBuffer to the PlayOutputBuffer, 
	inserting interpolated points or dropping occasional samples,
	as necessary for synchronization. This routine does not 
	increment the NextPlayBuffer. */
     /* returns # of free play buffers */
{
  int FreePlayBuffers;
  int Delta; /* 2* the number of extra free buffers */
  int InsertSpacing,RemoveSpacing;
  int PBCnt;
  int POBCnt;
  int Channel;
  static int UpdateCnt;

  FreePlayBuffers=(NextPlayBuffer-NextAvailBuffer+NumPlayBuffers) % NumPlayBuffers;
  UpdateCnt++;
  UpdateCnt %= 10; /* only update GUI every tenth time. */
  if (!UpdateCnt && panelinterfacewrite) {
    /* (low priority write) */
    qob_printf(panelinterfacewrite,0,"PlayBufPos=%d\n",(int)(FreePlayBuffers*100/NumPlayBuffers));
  }

  /*  if (PlayDebug) fprintf(stderr,"FreePlayBuffers=%ld\n",(long)FreePlayBuffers);*/



  
  /*Delta=FreePlayBuffers*2-NumPlayBuffers;*/
  Delta=0; /* this feature has been disabled */

 
  /* we ignore a Delta of 0, +1, or -1 */
  if (Delta > 1) {
    Delta--;
    /* At a delta of 2, gain of 1, insertspacing=500, or 1 in 500 points inserted. */
    
    InsertSpacing=100000/SyncGain/Delta;
    /*    fprintf(stderr,"InsertSpacing=%ld\n",(long)InsertSpacing);*/
    
    for (PBCnt=0,POBCnt=0;PBCnt < PlayBufferSize && POBCnt < POBSize;PBCnt++,POBCnt++) {
      if ((PBCnt%InsertSpacing)==0 && PBCnt > InsertPointThreshold && PlayBufferSize-PBCnt > InsertPointThreshold) {
	/* insert an extra point */
	for (Channel=0;Channel < PlayNumChannels;Channel++) {
	  PlayOutputBuffer[POBCnt*PlayNumChannels+Channel]=PlayBufferElem(NextPlayBuffer,PBCnt*PlayNumChannels+Channel); /*this does stairstep */
	  /*** IMPORTANT *** !!! BUG !!! ***
	    We should use some sort of interpolation to determine 
	    the inserted value. Perhaps 'deterministic estimation' from 
	    526. Time to review... */
	  
	}
	
	POBCnt++;
      }
      
      
      /* Copy data */
      if (POBCnt < POBSize) {
	for (Channel=0;Channel < PlayNumChannels;Channel++) {
	  PlayOutputBuffer[POBCnt*PlayNumChannels+Channel]=PlayBufferElem(NextPlayBuffer,PBCnt*PlayNumChannels+Channel);
	}
      }
      
    }
    
  }
  
  else if (Delta < -1) {
    Delta++;
    
    /* At a delta of 2, gain of 1, removespacing=500, or 1 in 500 points removed. */
    
    RemoveSpacing=-100000/SyncGain/Delta;

    /*    fprintf(stderr,"RemoveSpacing=%ld\n",(long)RemoveSpacing);*/

    for (PBCnt=0,POBCnt=0;PBCnt < PlayBufferSize;PBCnt++,POBCnt++) {

      if (PBCnt%RemoveSpacing==0 && PBCnt > RemovePointThreshold && PlayBufferSize-PBCnt > RemovePointThreshold) {
	/* remove a point from outgoing data */
	PBCnt++;
      }
      
      /* Copy data */
      for (Channel=0;Channel < PlayNumChannels;Channel++) {
	PlayOutputBuffer[POBCnt*PlayNumChannels+Channel]=PlayBufferElem(NextPlayBuffer,PBCnt*PlayNumChannels+Channel);
      }
      
      
    }
    
  } else {


    /* straight copy */
    for (PBCnt=0,POBCnt=0;PBCnt < PlayBufferSize;PBCnt++,POBCnt++) {
      /* Copy data */
      for (Channel=0;Channel < PlayNumChannels;Channel++) {
	PlayOutputBuffer[POBCnt*PlayNumChannels+Channel]=PlayBufferElem(NextPlayBuffer,PBCnt*PlayNumChannels+Channel);
      }
    }
  }

  /* set POB size data */
  POBBytes=POBCnt*2*PlayNumChannels;
  POBPos=0;

  return FreePlayBuffers;
  
}

int main(int argc, char *argv[])
{
  int Cnt,Cnt2;
  int GotMagic=0;
  fd_set readfd,writefd;
  int maxfd;
  float debugtime;
  int soundcard_ready_to_read=0,soundcard_ready_to_write=0;
  int PlayNextIsStatus;
  struct StatusBlk TmpStatus;

  struct StatusBlk *PlayStatus,*MOBStatus;
  struct DataBlkHeader *PlayHeader;

  int OldPlayUpsample;
  int OldPlayNumChannels;

  int FreePlayBuffers=0;
  int SelectShortCircuit=0;

  struct timeval tv;

  if (!geteuid()) {
    /* If we have root privlidges, set the nice level to
       a large negative number. Then relinquish privs. */
    
    nice(-20);

    if (getuid())
      setuid(getuid());
  }
  

  gargv=argv;

  for (Cnt=1; Cnt < argc; Cnt++) {
    if (!strncmp(argv[Cnt],"-magic",6)) {
      GotMagic=1;
    }
    else if (!strcmp(argv[Cnt],"-record")) {
      Mode=SIM_RECORD;
    }
    else if (!strcmp(argv[Cnt],"-play")) {
      Mode=SIM_PLAY;
    }
    else if (!strcmp(argv[Cnt],"-duplex")) {
      Mode=SIM_DUPLEX;
    }
    else if (!strcmp(argv[Cnt],"-nomonitor")) {
      NoMonitor=1;
    }
    else if (!strcmp(argv[Cnt],"-nopanel")) {
      NoPanel=1;
    }
    else if (!strncmp(argv[Cnt],"-title=",7)) {
      Title=argv[Cnt]+7;
    }
    else if (!strncmp(argv[Cnt],"-recordrate=",12)) {
      RecordSampleRate=strtol(argv[Cnt]+12,NULL,10);
    }
    else if (!strcmp(argv[Cnt],"-stereo")) {
      RecordNumChannels=2;
    } 
    else if (!strcmp(argv[Cnt],"-mono")) {
      RecordNumChannels=1;
    }
    else if (!strncmp(argv[Cnt],"-recordcardnum=",15)) {
      RecordCardNum=strtol(argv[Cnt]+15,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-recordcarddev=",15)) {
      RecordCardDev=strtol(argv[Cnt]+15,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-playcardnum=",13)) {
      PlayCardNum=strtol(argv[Cnt]+13,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-playcarddev=",13)) {
      PlayCardDev=strtol(argv[Cnt]+13,NULL,10);
    }
    else if (!strcmp(argv[Cnt],"-dac_impulse")) {
      SetDACMode(DAC_IMPULSE);
    }
    else if (!strcmp(argv[Cnt],"-dac_step")) {
      SetDACMode(DAC_STEP);
    }
    else if (!strcmp(argv[Cnt],"-dac_linear")) {
      SetDACMode(DAC_LINEAR);
    }
    else if (!strcmp(argv[Cnt],"-dac_filtered")) {
      SetDACMode(DAC_FILTERED);
    }
    else if (!strcmp(argv[Cnt],"-adc_unfiltered")) {
      SetADCMode(ADC_UNFILTERED);
    }
    else if (!strcmp(argv[Cnt],"-adc_filtered")) {
      SetADCMode(ADC_FILTERED);
    }
    else if (!strncmp(argv[Cnt],"-recordbuffersize=",18)) {
      RecordBufferSize=strtol(argv[Cnt]+18,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-numplaybuffers=",16)) {
      NumPlayBuffers=strtol(argv[Cnt]+16,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-recordmaxrate=",15)) {
      RecordMaxRate=strtol(argv[Cnt]+15,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-playmaxrate=",13)) {
      PlayMaxRate=strtol(argv[Cnt]+13,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-recordhwfragsize=",18)) {
      recordlog2fragsize=(int)(log(strtol(argv[Cnt]+18,NULL,10))/log(2.0) +.01);
    }
    else if (!strncmp(argv[Cnt],"-playhwfragsize=",16)) {
      playlog2fragsize=(int)(log(strtol(argv[Cnt]+16,NULL,10))/log(2.0) +.01);
    }
    else if (!strncmp(argv[Cnt],"-recordhwnumfrags=",18)) {
      recordlog2numfrags=(int)(log(strtol(argv[Cnt]+18,NULL,10))/log(2.0) +.01);
    }
    else if (!strncmp(argv[Cnt],"-playhwnumfrags=",16)) {
      playlog2numfrags=(int)(log(strtol(argv[Cnt]+16,NULL,10))/log(2.0) +.01);
    }
    /* else if (!strncmp(argv[Cnt],"-recordstatusperiod=",20)) {
      RecordStatusPeriod=strtol(argv[Cnt]+20,NULL,10);
      }*/
    else if (!strcmp(argv[Cnt],"-recorddebug")) {
      RecordDebug=1;
    }
    else if (!strcmp(argv[Cnt],"-playdebug")) {
      PlayDebug=1;
    }
    else if (!strcmp(argv[Cnt],"-drainsync")) {
      DrainSync=1;
    }
    
  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:   #0     Input to playback unit\n"
	    "         #1     Output from record unit\n"
	    "         #2     Error messages (do not connect)\n"
	    "         #3     Monitor output\n"
	    "\n"
	    "Usage;\n"
	    /*	    "   %s -magic -record -play -duplex -nomonitor\n"
		    "    -title=window_title\n"
		    "    -recordrate=%d -stereo -mono -cardnum=0 -recordmaxrate=%d\n"
		    "    -playmaxrate=%d -recordhwfragsize=%d -playhwfragsize=%d\n"
		    "    -recordhwnumfrags=%d -playhwnumfrags=%d -dac_impulse\n"
		    "    -dac_step -dac_linear -dac_filtered -adc_unfiltered\n"
		    "    -adc_filtered -recordbuffersize=%d -recordstatusperiod=%d\n"
		    "    -numplaybuffers=%d",argv[0],(int)RecordSampleRate); */
	    "   %s -magic -record -play -duplex -nomonitor\n"
	    "   -nopanel -title=%s -recordrate=%d -stereo -mono\n"
	    "   -recordcardnum=%d -recordcarddev=%d -playcardnum=%d\n"
	    "   -playcarddev=%d -dac_impulse\n"
	    "   -dac_step -dac_linear -dac_filtered -adc_unfiltered\n"
	    "   -adc_filtered -recordbuffersize=%d -numplaybuffers=%d\n"
	    "   -recordmaxrate=%d -playmaxrate=%d -recordhwfragsize=%d\n"
	    "   -playhwfragsize=%d -recordhwnumfrags=%d -playhwnumfrags=%d\n"
	    "   -recorddebug -playdebug -drainsync\n"
	    ,argv[0],Title,(int)RecordSampleRate,(int)RecordCardNum,(int)RecordCardDev,(int)PlayCardNum,(int)PlayCardDev,(int)RecordBufferSize,
	    (int)NumPlayBuffers,(int)RecordMaxRate,(int)PlayMaxRate,
	    (recordlog2fragsize < 0) ? ((int)-1):((int)(1<<recordlog2fragsize)),
	    (playlog2fragsize < 0) ? ((int)-1):((int)(1<<playlog2fragsize)),
	    (recordlog2numfrags < 0) ? ((int)-1):((int)(1<<recordlog2numfrags)),
	    (playlog2numfrags < 0) ? ((int)-1):((int)(1<<playlog2numfrags)));
    exit(1);
  }
    
  if (!NoPanel) {
    panelinterfacereadfd=initpanelinterface(argv[0],&panelinterfacewrite);
  }
  
  /* open devices */
  if (Mode==SIM_DUPLEX && RecordCardNum==PlayCardNum
#ifdef LINUX_ALSA /* under ALSA, proper duplex mode doesn't seem to work,
		     but everything's fine if we open the cards separately */
      && 0 && RecordCardDev==PlayCardDev
#endif 
      ) {
    /* one audio device, duplex mode */
    PlayMode=RecordMode=SIM_DUPLEX;
    recordsoundcard=sound_init(RecordCardNum,RecordCardDev,Mode,recordlog2fragsize,recordlog2numfrags);
    playsoundcard=recordsoundcard;
  } else if (Mode==SIM_DUPLEX) {
    /* two audio devices: record and play */
    RecordMode=SIM_RECORD;
    recordsoundcard=sound_init(RecordCardNum,RecordCardDev,SIM_RECORD,recordlog2fragsize,recordlog2numfrags);
    PlayMode=SIM_PLAY;
    playsoundcard=sound_init(PlayCardNum,PlayCardDev,SIM_PLAY,playlog2fragsize,playlog2numfrags);
  } else if (Mode==SIM_RECORD) {
    /* single audio device: record */
    RecordMode=SIM_RECORD;
    recordsoundcard=sound_init(RecordCardNum,RecordCardDev,SIM_RECORD,recordlog2fragsize,recordlog2numfrags);
  } else if (Mode==SIM_PLAY) {
    PlayMode=SIM_PLAY;
    playsoundcard=sound_init(PlayCardNum,PlayCardDev,SIM_PLAY,playlog2fragsize,playlog2numfrags);
  }

  if (((Mode==SIM_RECORD || Mode==SIM_DUPLEX) && recordsoundcard->recordsoundfh < 0) || ((Mode==SIM_PLAY || Mode==SIM_DUPLEX) && playsoundcard->playsoundfh < 0)) {
    fprintf(stderr,"%s: Unable to open soundcard\n",argv[0]);
    exit(1);
  }
  
  

  /* nonblock mode is done in the driver now
 if (playsoundcard && playsoundcard->playsoundfh >= 0)
    setnonblocking(playsoundcard->playsoundfh);
  if (recordsoundcard && recordsoundcard->recordsoundfh >= 0)
    setnonblocking(recordsoundcard->recordsoundfh);
  */
  if (Mode==SIM_RECORD || Mode==SIM_DUPLEX) 
    setnonblocking(RecordPipe);
  if (Mode==SIM_PLAY || Mode==SIM_DUPLEX)
    setnonblocking(PlayPipe);

  if (!NoMonitor)
    setnonblocking(MonitorPipe);
  
  if (Mode==SIM_RECORD || Mode==SIM_DUPLEX) {
    /* Configure rate now */
    CalcRecordRates(RecordSampleRate,RecordNumChannels);

    /* build buffers */
    BuildRecordBuffers(RecordBufferSize);
    /* prepare RecordBuffer for input */
    CurRecordPos=0;    

    /* Transmit initial status block */
    WriteRecordStatus(); /* send status to output pipe */
    RecordStatusPromised=0;
    WriteRTBuf();


  }

  /* Prepare play system to get configuration block */
  PRBBytes=sizeof(struct StatusBlk);
  PRBPos=0;
  PlayReceiveBuf=malloc(PRBBytes);
  PlayConfigured=0;
  
  /* read-write loop */
  
  for(;;) {
    maxfd=maxva(RecordPipe,PlayPipe,(recordsoundcard==NULL) ? -1:recordsoundcard->recordsoundfh,(playsoundcard==NULL) ? -1:playsoundcard->playsoundfh,MonitorPipe,panelinterfacereadfd,(panelinterfacewrite==NULL) ? -1:panelinterfacewrite->writefd,-2);
    FD_ZERO(&readfd);
    FD_ZERO(&writefd);
    
    SelectShortCircuit=0;

    if (Mode==SIM_DUPLEX || Mode==SIM_RECORD) {
      if (RTPos < RTBytes) { /* if have data to send on record pipe */
	FD_SET(RecordPipe,&writefd);
      }
      if (CurRecordPos < RecordBufferBytes) { /* If can accept data from soundcard */
	FD_SET(recordsoundcard->recordsoundfh,&readfd);
      }
      
    }
    
    if (Mode==SIM_DUPLEX || Mode==SIM_PLAY) {
      if (DrainSync) {
	/* in DrainSync mode we only allow ourselves to read from PRB if
	   there is PlayBuffer space available for it */
	if ((PRBPos < PRBBytes) &&  ((NextAvailBuffer + 1) % NumPlayBuffers != NextPlayBuffer)) { /* only allow read if we won't overflow */
	  FD_SET(PlayPipe,&readfd);
	}

      } else {
	if (PRBPos < PRBBytes) { /* if space to read from incoming play buffer */
	  FD_SET(PlayPipe,&readfd);
	}
      }
      
      /* short circuit the select if there's data available in the PRB */
      if (PRBBytes && PRBPos==PRBBytes)
	SelectShortCircuit=1;
      
      if (POBPos < POBBytes && POBBytes && PlayConfigured) {
	/* if ready to write to soundcard */
	/* fprintf(stderr,"setting playsoundcard->playsoundfh bit\n"); */
	FD_SET(playsoundcard->playsoundfh,&writefd);
      }
    }
    
    if (!NoMonitor && MOBPos < MOBBytes && PlayConfigured) {
      FD_SET(MonitorPipe,&writefd);
    }
    
    if (!NoPanel) {
      if (panelinterfacereadfd >= 0) {
	FD_SET(panelinterfacereadfd,&readfd);
      }
      if (panelinterfacewrite && panelinterfacewrite->writefd >= 0 && queued_outbuffer_data_available(panelinterfacewrite)) {
	FD_SET(panelinterfacewrite->writefd,&writefd);
      }
    }

    tv.tv_sec=0;
    tv.tv_usec=selecttimeout;
    
    
#ifdef LINUX_OSS_DRIVER_BUG_HACK
    /* The Linux OSS driver, at least with some soundcards, has a bug
       where the select() call may wait even though there is buffer space
       available. We work around this by tracking whether or not there
       was buffer space available as of the last read/write with the
       soundcard_ready_to_read and soundcard_ready_to_write variables.
       Here, we check the status of those variables, and if appropriate
       change the select() to a zero timeout and force the bits on
       as appropriate. */

    {
      int gotsoundcardread=0,gotsoundcardwrite=0;
      
      if (soundcard_ready_to_read && FD_ISSET(recordsoundcard->recordsoundfh,&readfd)) {
	tv.tv_sec=0;
	tv.tv_usec=0;
	gotsoundcardread=1;
      }
      if (soundcard_ready_to_write && FD_ISSET(playsoundcard->playsoundfh,&writefd)) {
	tv.tv_sec=0;
	tv.tv_usec=0;
	gotsoundcardwrite=1;
      }
      
      
#endif /* LINUX_OSS_DRIVER_BUG_HACK */      
      if (SelectShortCircuit) { /* not currently used */
	/* force select() to return immediately */
	tv.tv_sec=0;
	tv.tv_usec=0;
      }
      select(maxfd+1,&readfd,&writefd,NULL,&tv);
#ifdef LINUX_OSS_DRIVER_BUG_HACK
      if (gotsoundcardread) FD_SET(recordsoundcard->recordsoundfh,&readfd);
      if (gotsoundcardwrite) FD_SET(playsoundcard->playsoundfh,&writefd);
      
    }

#endif /* LINUX_OSS_DRIVER_BUG_HACK */
    
    
    /* read/write the available data */
    if (Mode==SIM_RECORD || Mode==SIM_DUPLEX) {
      if (FD_ISSET(RecordPipe,&writefd)) {
	WriteRTBuf();
     }
      if (FD_ISSET(recordsoundcard->recordsoundfh,&readfd)) {
	/* Read data from soundcard */
	CurRecordPos=sound_readbuf(recordsoundcard,RecordBuffer,RecordBufferBytes,CurRecordPos,&soundcard_ready_to_read,RecordDebug);
      }
    }
    if (Mode==SIM_PLAY || Mode==SIM_DUPLEX) {
      if (FD_ISSET(PlayPipe,&readfd)) {
	ReadPRBuf();
      }

      if (
#ifdef LINUX_OSS_DRIVER_BUG2_HACK
	  POBPos < POBBytes && POBBytes
#else /* LINUX_OSS_DRIVER_BUG2_HACK */
	  FD_ISSET(playsoundcard->playsoundfh,&writefd)
#endif /* LINUX_OSS_DRIVER_BUG2_HACK */
	  ) {
	POBPos=sound_writebuf(playsoundcard,PlayOutputBuffer,POBBytes,POBPos,&soundcard_ready_to_write,PlayDebug);
      }
      if (!NoMonitor && FD_ISSET(MonitorPipe,&writefd)) {
	WriteMOBuf();
      }
    }
    /* Process data from control panel */
    if (!NoPanel && panelinterfacereadfd >= 0 && FD_ISSET(panelinterfacereadfd,&readfd)) {
      pi_readchars();
    }
    if (!NoPanel && panelinterfacewrite && FD_ISSET(panelinterfacewrite->writefd,&writefd)) {
      write_queued_outbuffer(panelinterfacewrite);
    }



    /* Now try to transfer data between input/output buffers.
       
       The flow is as follows: 

          Recording Section
          -----------------

	  Soundcard -> RecordBuffer -> Downsample -> RecordTransmitBuffer -> Pipe # 1

	  
	  Playing Section
	  ---------------
	    
	  Pipe #0 -> PlayReceiveBuffer -> Upsample -> PlayBuffers -> PlayOutputBuffer -> Soundcard
                                                   \-> MonitorOutputBuf -> Pipe #3
*/
    
    /* Move data for recording section */

    if (Mode==SIM_DUPLEX || Mode==SIM_RECORD) {
      /* reset output buffer if empty */
      if (RTBytes > 0 && RTPos==RTBytes) {
	if (RecordStatusPromised) {
	  /* Send status message */
	  if (QRP.Valid) {
	    /* This is a good time update the RecordParams with the queued values */
	    
	    CalcRecordRates(QRP.NewRecordRate,QRP.NewNumChannels);
	    BuildRecordBuffers(QRP.NewBufSize);
	    QRP.Valid=0;
	  }
	  RecordStatusPromised=0;

	  WriteRecordStatus();
	  WriteRTBuf(); /* Write it if possible */
	  if (RTPos==RTBytes) RTPos=RTBytes=0; /* empty if we did the write */
	} else {
	  
	  RTBytes=0;
	  RTPos=0;	
	}
	
      }
      
      /* if RecordBuffer is full & RTBuffer is empty, then we can empty 
	 RecordBuffer into RTBuffer by performing downsampling and 
	 filtering if necessary. Let's not forget to add the header information
	 to RTBuffer! */
      
      if (RecordBufferBytes != 0 && RecordBufferBytes==CurRecordPos && RTBytes==0) {
	RecordFrameNumber++;
	DownsampleRecordBuffer();

      }
      
    }

    /* Now do the transfers shown in the above comment for the playing 
       section */

    if (Mode==SIM_DUPLEX || Mode==SIM_PLAY) {
      
      if (PlayConfigured) {
	/* Check to see if the MonitorOutputBuf is fully sent -> empty it */
	if (MOBBytes > 0 && MOBPos==MOBBytes) MOBBytes=MOBPos=0;
	
	/* Check to see if the PlayOutputBuffer is fully sent -> empty it */
	if (POBBytes > 0 && POBPos==POBBytes) POBBytes=POBPos=0;
	
	/* See if we can move a block from the PlayBuffers to the 
	   PlayOutputBuffer */
	if (FPBTarget < 0 && POBBytes==0 && NextPlayBuffer!= NextAvailBuffer) {
	  CopyPBtoPOB(NextPlayBuffer);
	  /*	  if (PlayDebug) fprintf(stderr,"PB #%d emptied\n",NextPlayBuffer);*/
	  /* Go on to the next play buffer */
	  NextPlayBuffer++;
	  if (NextPlayBuffer==NumPlayBuffers) NextPlayBuffer=0;
	} else if (FPBTarget < 0 && NextPlayBuffer==NextAvailBuffer) {
	  /* no data left in our buffer. Let it fill up before continuing */
	  FPBTarget=NumPlayBuffers/2;
#ifdef DEBUG_BUFFERING
	  if (PlayDebug) fprintf(stderr,"Letting buffer fill.\n");
#endif
	}
      }
      

      

      
      /* Check to see if the PlayReceiveBuffer is full and the 
	 MonitorOutputBuf empty. 
	 If it is, perform upsampling and transfer the data from the 
	 PlayReceiveBuffer to both
	 the MonitorOutputBuf and the next PlayBuffer. If the next PlayBuffer
	 is the next one to be played, do not transfer the data to the 
	 PlayBuffer (just throw it out). 

	 Also, we should be aware that some of the blocks received in the 
	 PlayReceiveBuffer are status blocks. These are not passed on to 
	 the PlayBuffers, but their contents are used to update parameters 
	 and they are passed to the MonitorOutputBuf. */
      if (MOBBytes==0 && PRBBytes != 0 && PRBPos==PRBBytes) {
	PlayStatus=(struct StatusBlk *)PlayReceiveBuf;
	if (PlayStatus->MagicByte[3]==((UINT8)0x7f)) {
	  /* Status block */
	  assert(PRBBytes==sizeof(struct StatusBlk));

	  /*	  fprintf(stderr,"Ready to configure playing\n");*/
	  if (!PlayConfigured) {
	    /* we haven't configured playing yet, but we just got our 
	       status block, so we configure now. */
	    PlayReceiveBuf=NULL; /* Prevent BuildPlayBuffers() from freeing
				    our PlayStatus block */
	    CalcPlayRates(PlayStatus->SampleRate,PlayStatus->NumChannels);
	    BuildPlayBuffers(PlayStatus->BlockSize,NumPlayBuffers);

	    /* Stuff the original data back into the newly allocated 
	       PlayReceiveBuf */
	    memcpy(PlayReceiveBuf,PlayStatus,sizeof(struct StatusBlk));
	    PRBBytes=sizeof(struct StatusBlk);
	    PRBPos=PRBBytes;
	    /* Free the original PlayReceiveBuf */
	    free(PlayStatus);
	    PlayStatus=(struct StatusBlk *)PlayReceiveBuf;
	    PlayConfigured=1;
	  }


	  /* See if we have to change our internal status */
	  OldPlayUpsample=PlayUpSample;
	  OldPlayNumChannels=PlayNumChannels;
	  if (PlayStatus->SampleRate != DesiredPlaySampleRate || PlayStatus->NumChannels != PlayNumChannels) {
	    CalcPlayRates(PlayStatus->SampleRate,PlayStatus->NumChannels);
	  }

	  /* copy the status block to the MonitorOutputBuf */
	  if (!NoMonitor) {
	    memcpy(&TmpStatus,PlayReceiveBuf,sizeof(struct StatusBlk));
	  }
	  
	  if (PlayNumChannels != OldPlayNumChannels || OldPlayUpsample != PlayUpSample || PlayStatus->BlockSize != PRBSize) {
	    /* WARNING: BuildPlayBuffers() reallocates the PlayBuffers, the PlayReceiveBuf, and the PlayOutputBuffer, so any data there is LOST */
	    BuildPlayBuffers(PlayStatus->BlockSize,NumPlayBuffers);
	  }
	  
	  if (!NoMonitor) {
	    if (MOBPromisedStatusBlk) { /* Can't send a status block unless
					   we told them we were going to
					   send one */
	      
	      
	      memcpy(MonitorOutputBuf,&TmpStatus,sizeof(struct StatusBlk));
	      
	      /* The MonitorOutput SampleRate is PlayUpSample*the nominal
		 sample rate, and the BlockSize is also PlayUpSample*nominal.
		 */
	      
	      MOBStatus=(struct StatusBlk *) MonitorOutputBuf;
	      /*fprintf(stderr,"PlayHWSampleRate=%d\n",(int)PlayHWSampleRate); */
	      MOBStatus->SampleRate=PlayHWSampleRate;
	      MOBStatus->BlockSize=PlayBufferSize;
	      
	      MOBBytes=sizeof(struct StatusBlk);
	      MOBPromisedStatusBlk=0;
	    }
	  }
	  

	  
	  /* Free up the PlayReceiveBuffer*/
	  PRBPos=0;
	  PRBBytes=PRBSize*2*PlayNumChannels+sizeof(struct DataBlkHeader);
	} else {
	  /* Got a DataBlk */
	  /* Pass it on to MonitorOutputBuffer*/
	  /* Upsample, filter if necessary, data from PlayReceiveBuffer 
	     to the PlayBuffers and MonitorOutputBuf */

	  
	  UpsamplePlayBuffer();

	  FreePlayBuffers=(NextPlayBuffer-NextAvailBuffer+NumPlayBuffers) % NumPlayBuffers;
	  if (FreePlayBuffers <= FPBTarget) FPBTarget=-1;

	  PRBPos=0;
	  PlayHeader=(struct DataBlkHeader *)PlayReceiveBuf;

	  PlayNextIsStatus=PlayHeader->NextBlockIsStatusBlk;
	  
	  if (QPP.Valid) {
	    /* Hold on. We have a queued parameter change. Let's do it now */
	    BuildPlayBuffers(PlayBufferSize/PlayUpSample,QPP.NewNumBuffers); /* This call destroys all play buffers */
	    QPP.Valid=0;
	  }
	  
	  if (PlayNextIsStatus) PRBBytes=sizeof(struct StatusBlk);
	  else PRBBytes=PRBSize*2*PlayNumChannels+sizeof(struct DataBlkHeader);
	  
	}
	
      }
      
      if (PlayOutputBuffer && FPBTarget >= 0 && POBBytes==0) {
	/* If we're waiting for the play buffer to fill up, while 
	   we're waiting, just play zeros */

#ifdef DEBUG_BUFFERING
	fprintf(stderr,"Padding with zeros\n");
#endif

	POBBytes=PlayBufferSize*2*PlayNumChannels;
	for (Cnt=0;Cnt < PlayBufferSize;Cnt++) {
	  for (Cnt2=0;Cnt2 < PlayNumChannels;Cnt2++) {
	    PlayOutputBuffer[Cnt*PlayNumChannels+Cnt2]=0;
	  }
	}
	
	POBPos=0;
      }

      
    }
    
    
  }
  
  return 0;
}


