/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* syncswitch.c --  Allow a user to switch back and forth between 
                    differently processed versions of the same source.
		    We require synchronization of frame numbers between
		    two sources. */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "io_spec_old.h"
#include "oldbuffer.h"
#include "linklist.h"
#include "util.h"
#include "syncswitch.h" 

/* Notes:  If jump-back mode is enabled this module requires  that the 
   synchronization source be the signal destination, not the source. 
   (the opposite convention from most of Aglaophone). This precludes using
   real-time digitized input when jump-back is enabled. The recommended
   connection mode for this module with jump-back is:
 non-rate-limited source   -> Process A --> blindswitch --> pacer --> out to
                    \-------> Process B -----/                      soundcard

   in the future, a "drain sync" mode may be available in soundcardio
   which would eliminate the need for the pacer module if that option is set.

   We currently require that the two inputs have the same blocksize, 
   sampling rate, and number of channels

   */

#define MODINC(number,base) (((number)+1+(base))%(base))

int InputAPipe=0;
int InputBPipe=3;

int OutputPipe=1;

int GUIRead=-1; /* Use this pipe for reading from the GUI */
struct queued_outbuffer *GUIWrite=NULL; /* Use this buffer for writing to the GUI */

/* waveform parameters */
int NumChannels;
int SampleRate;
int BlockSize;


struct BufferSet {
  int NumSecs; /* number of seconds to buffer not meaningful for the FreeBuf */
  int NumBufs; /* size of array */
  int NextToFill;
  int NextToEmpty;
  struct SBuf **Bufs; 
};


struct BufferSet BufsA; /* Buffering of input A */
struct BufferSet BufsB; /* Buffering of input B */
int NextToOutput; /* next buffer to transmit */

int BufferSecs=5;
int BufferBufs;
int JumpBackSecs=3;
int JumpBackBufs;

int Switch='A'; /* Should be 'A' or 'B' */
int JumpBack=1; /* is jump back mode currently on or off (as opposed to 
		   disabled) */

/* These are pointers to specific buffers in the BufferSets 
   (or NULL) */
struct SBuf *ReadBufA=NULL;
struct SBuf *ReadBufB=NULL;
struct SBuf *WriteBuf=NULL;

int NextAIsStatus=0; /* Set if next block to be read on input A should be a 
			status block */
int NextBIsStatus=0; /* Set if next block to be read on input B should be a 
			status block */
struct SBuf *WriteStatusBuf=NULL; /* Used only for transmitting status blocks */
int PromisedStatus=0;
int WriteFrameNum=0;
/*int StatusBlkSpacing=100;*/
int StatusBlkSpacing=1; /* setting the status block spacing to 1 makes us 
			   compatible with the new io_spec */

int NumAThrowaway=0; /* Number of proceeding blocks (not including status) 
			on A to discard as we resynchronize following a
			dynamic parameter change */
int NumBThrowaway=0;


/* String (and other) parameters to be passed to panel interface */
char *InputLabelA="Input A";
char *InputLabelB="Input B";

char *BlindLabel0="Input 1";
char *BlindLabel1="Input 2";

char *Message="Choose between the two inputs";
char *RevealMessageA="You selected input A";
char *RevealMessageB="You selected input B";

int NoJumpBack=0; /* is jump-back mode disabled (as opposed to off)? */
int Blind=0; /* Are we in blind-test mode? */


/* Should add this to buffer.c: */
void SBufEmpty(struct SBuf *Buf)
{
  Buf->Bytes=0;
  Buf->Pos=0;
}

/*** IMPORTANT *** Must patch buffer.c: SBufIsReady() to 
  return (Buf->Bytes > 0 && Buf->Pos < Buf->Bytes) **/


int NumInUse(struct BufferSet *Set)
{
  return (Set->NextToFill-Set->NextToEmpty+Set->NumBufs)%Set->NumBufs;
}

void InitBufferSet(struct BufferSet *Set,int Secs,int Num)
{
  int Cnt;

  Set->NextToFill=0;
  Set->NextToEmpty=0;

  if (Set->NumBufs==Num) {
    /* no change in # of buffers -- no initialization needed */
    return;
  }

  /* Otherwise, we clear what's there, and reallocate */
  if (Set->NumBufs && Set->Bufs) {
    for (Cnt=0;Cnt < Set->NumBufs;Cnt++) {
      DeleteSBuf(Set->Bufs[Cnt]);
      Set->Bufs[Cnt]=NULL;
    }
    free(Set->Bufs);
    Set->Bufs=NULL;
  }
  
  Set->NumSecs=Secs;
  Set->NumBufs=Num;
  Set->Bufs=malloc(sizeof(struct Sbuf *)*Set->NumBufs);
  for (Cnt=0;Cnt < Set->NumBufs;Cnt++) {
    Set->Bufs[Cnt]=CreateSBuf();
  }
}

void AllocBufferSet(struct BufferSet *Set)
     /* Reallocate the physical buffers within a buffer set to correspond
	with the current # of channels and block size */
{
  int Cnt;
  
  for (Cnt=0;Cnt < Set->NumBufs;Cnt++) {
    Set->Bufs[Cnt]->NumChannels=NumChannels;
    SBufNewSize(Set->Bufs[Cnt],BlockSize);
  }
}

void RebuildBuffers(void)
     /* reallocate all buffers to new size, according to SampleRate,
        NumChannels, and BlockSize. */
{
  int Secs;

  Secs=BufferSecs+JumpBackSecs;
  BufferBufs=BufferSecs*SampleRate/BlockSize+1;
  JumpBackBufs=JumpBackSecs*SampleRate/BlockSize+1;

  InitBufferSet(&BufsA,Secs,BufferBufs+JumpBackBufs);
  AllocBufferSet(&BufsA);

  InitBufferSet(&BufsB,Secs,BufferBufs+JumpBackBufs);
  AllocBufferSet(&BufsB);

  NextToOutput=0;
  
  
  ReadBufA=NULL;
  ReadBufB=NULL;
  WriteBuf=NULL;

}

int BuffDiff(int Larger,int Smaller,int Modulo)
     /* Calculates the difference Larger-Smaller, 
	based on the assumption that Larger is beyond (or equal to) Smaller */
{
  return (Larger-Smaller+Modulo)%Modulo;
}

void PostStatus(struct SBuf *StatusBuf)
{
  struct StatusBlk *Status;
  int NumA,NumB;

  Status=SBufGetStatus(StatusBuf);
  

  if ((Status->NumChannels != NumChannels) ||
      (Status->SampleRate != SampleRate) ||
      (Status->BlockSize != BlockSize)) {
    NumChannels=Status->NumChannels;
    SampleRate=Status->SampleRate;
    BlockSize=Status->BlockSize;
    WriteFrameNum=Status->FrameNumber;
  
    /* These are the number of recorded buffers in A and B */
    NumA=BuffDiff(BufsA.NextToFill,BufsA.NextToEmpty,BufsA.NumBufs);
    NumB=BuffDiff(BufsB.NextToFill,BufsB.NextToEmpty,BufsB.NumBufs);
    if (NumA > NumB) {
      NumBThrowaway+=NumA-NumB;
    }
    if (NumA < NumB) {
      NumAThrowaway+=NumB-NumA;
    }

    /* Major parameter change */
    RebuildBuffers();
    /* We need to send a status block. But we can't unless we promised one.
       So, if we didn't promise one, we send a block of zeros instead. */
    if (!PromisedStatus) {
      assert(!BufsA.NextToFill);
      assert(!BufsB.NextToFill);
      BufsA.NextToFill++;
      BufsB.NextToFill++;
      PromisedStatus=1;
      SBufSetDataHeader(BufsA.Bufs[0],PromisedStatus);
      memset(BufsA.Bufs[0]->Data+sizeof(struct DataBlkHeader)/sizeof(INT16),0,sizeof(INT16)*BlockSize*NumChannels);
      SBufSetDataHeader(BufsB.Bufs[0],PromisedStatus);
      memset(BufsB.Bufs[0]->Data+sizeof(struct DataBlkHeader)/sizeof(INT16),0,sizeof(INT16)*BlockSize*NumChannels);
      
    }
  }
}

int IsBufSpace(struct BufferSet *Set)
     /* Returns non-zero if there is an empty buffer available */
{
  return MODINC(Set->NextToFill,Set->NumBufs)!=Set->NextToEmpty;
}



void SwapIndexedBuf(struct BufferSet *Set,int *Index,struct SBuf *NewData)
     /* Swap SBuf into spot in Set specified by *Index, Perform 
	modulo-n increment of *Index */
{
  struct SBuf Tmp;
  
  Tmp=*Set->Bufs[*Index];
  *Set->Bufs[*Index]=*NewData;
  *NewData=Tmp;
  
  *Index=MODINC(*Index,Set->NumBufs);

  
}


int SwapNextBuf(struct BufferSet *Set,struct SBuf *NewData)
     /* Swap SBuf into next slot in Set if there is space
	returns 1 on success, 0 if there was no space */
{
  /* swap the data from NewData into the next available buffer of Set */

  if (!IsBufSpace(Set)) return 0;
  
  SwapIndexedBuf(Set,&Set->NextToFill,NewData);

  return 1;
}


int InActiveRegion(struct BufferSet *Set, int Num)
     /* Determines if buffer Num is >= NextToEmpty and < NextToFill */
{
  int RelNum,RelFill;

  RelNum=(Num-Set->NextToEmpty+Set->NumBufs)%Set->NumBufs;
  RelFill=(Set->NextToFill-Set->NextToEmpty+Set->NumBufs)%Set->NumBufs;
  return RelNum >= 0 && RelNum < RelFill; 
}


void PerformSwitch(int Position)
     /* assumes that a switch is occuring (i.e. new Position is different
	from old position */
{
  if (Position != Switch) {
    assert(BufsA.NextToEmpty==BufsB.NextToEmpty);
    Switch=Position;
    if (!NoJumpBack && JumpBack) {
      NextToOutput=BufsA.NextToEmpty; /* Rewind as far as we can go */
    }
  }
}

int main(int argc, char *argv[]) 
{
  int GotMagic=0;
  struct StatusBlk *StatusA,*StatusB;
  fd_set readfd,writefd;
  int maxfd;
  int Cnt;
  
  for (Cnt=1; argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    if (!strncmp(argv[Cnt],"-LabelA=",8)) {
      InputLabelA=argv[Cnt]+8;
    }
    if (!strncmp(argv[Cnt],"-LabelB=",8)) {
      InputLabelB=argv[Cnt]+8;      
    }
    if (!strcmp(argv[Cnt],"-nojumpback")) {
      NoJumpBack=1;
    }
    if (!strcmp(argv[Cnt],"-blind")) {
      Blind=1;
    }
    if (!strncmp(argv[Cnt],"-Blind1=",8)) {
      BlindLabel0=argv[Cnt]+8; 
    }
    if (!strncmp(argv[Cnt],"-Blind2=",8)) {
      BlindLabel1=argv[Cnt]+8; 
    }
    if (!strncmp(argv[Cnt],"-Message=",9)) {
      Message=argv[Cnt]+9; 
    }
    if (!strncmp(argv[Cnt],"-RevealA=",9)) {
      RevealMessageA=argv[Cnt]+9; 
    }
    if (!strncmp(argv[Cnt],"-RevealB=",9)) {
      RevealMessageB=argv[Cnt]+9; 
    }
    if (!strncmp(argv[Cnt],"-buffersize=",12)) {
      BufferSecs=strtol(argv[Cnt]+12,NULL,10);
    }
    if (!strncmp(argv[Cnt],"-jumpbacksize=",14)) {
      JumpBackSecs=strtol(argv[Cnt]+14,NULL,10);
    }
    
  }
  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   input stream A\n"
	    "        #1   output stream\n"
	    "        #2   error messages (do not connect)\n"
	    "        #3   input stream B\n"
	    "\n"
	    "Usage:\n"
	    "   %s -magic \"-LabelA=%s\" \"-LabelB=%s\"\n"
	    "   -nojumpback -blind\n"
	    "   \"-Blind1=%s\" \"-Blind2=%s\" \"-Message=%s\""
	    "   \"-RevealA=%s\" \"-RevealB=%s\""
	    "   -buffersize=%d -jumpbacksize=%d\n"
	    "\n"
	    "  -LabelA=%s -- set the result label corresponding to input A\n"
	    "  -LabelB=%s -- set the result label corresponding to input A\n"
	    "  -nojumpback -- disable the jump-back feature. This is required\n"
	    "                 if the signal chain is source-synchronized (the\n"
	    "                 Aglaophone convention)\n"
	    "  -blind      -- use the blindswitchpanel.java GUI for blind\n"
	    "                 comparison of processing paths\n"
	    "  -Blind1=%s -- Specify label for blind input #1\n"
	    "  -Blind2=%s -- Specify label for blind input #2\n"
	    "  -Message=%s -- Specify greeting/instructions on bottom of window\n"
	    "  -RevealA=%s -- Specify bottom of window message if the selected\n"
	    "  blind input is determined to be input A\n"
	    "  -RevealB=%s -- Specify bottom of window message if the selected\n"
	    "  blind input is determined to be input B\n"
	    "  -buffersize=%d -- Specify the maximum buffering between the two\n"
	    "  inputs (in seconds)\n"
	    "  -jumpbacksize=%d -- Specify the amount to jump back, in seconds\n"
	    ,argv[0],InputLabelA,InputLabelB,BlindLabel0,BlindLabel1,Message,RevealMessageA,RevealMessageB,(int)BufferSecs,(int)JumpBackSecs,InputLabelA,InputLabelB,BlindLabel0,BlindLabel1,Message,RevealMessageA,RevealMessageB,(int)BufferSecs,(int)JumpBackSecs);
    exit(1);
    
  }

  /* Get the GUI started */

  GUIRead=initsspi(argv[0],&GUIWrite);

  /* Build buffers: */
  /* first a little temporary one for the initial status block */
  ReadBufA=CreateSBuf();
  SBufPrepareStatus(ReadBufA);
  while (!ReadSBuf(ReadBufA,InputAPipe));
  StatusA=SBufGetStatus(ReadBufA);
  if (!SBufCheckStatus(ReadBufA)) {
    fprintf(stderr,"%s: Error in input A initial status block\n");
    exit(1);
  }

  ReadBufB=CreateSBuf();
  SBufPrepareStatus(ReadBufB);
  while (!ReadSBuf(ReadBufB,InputBPipe));
  StatusB=SBufGetStatus(ReadBufB);
  if (!SBufCheckStatus(ReadBufB)) {
    fprintf(stderr,"%s: Error in input B initial status block\n",argv[0]);
    exit(1);
  }
  
  if ((StatusA->NumChannels != StatusB->NumChannels) ||
      (StatusA->SampleRate != StatusB->SampleRate) || 
      (StatusA->FrameNumber != StatusB->FrameNumber) || 
      (StatusA->BlockSize != StatusB->BlockSize)) {
    fprintf(stderr,"%s: Parameter match error between inputs A and B\n",argv[0]);
    exit(1);
  }
  
  NumChannels=StatusA->NumChannels;
  SampleRate=StatusA->SampleRate;
  BlockSize=StatusA->BlockSize;
  WriteFrameNum=StatusA->FrameNumber;
  
  /* Now that we know # of channels, sample rate, and block size,
     we can allocate the buffers. Note that StatusA and StatusB are not 
     meaningful pointers after this call. */
  DeleteSBuf(ReadBufA);
  DeleteSBuf(ReadBufB);

  WriteStatusBuf=CreateSBuf();
  
  RebuildBuffers();
  
  setnonblocking(InputAPipe);
  setnonblocking(InputBPipe);
  setnonblocking(OutputPipe);

  ReadBufA=BufsA.Bufs[BufsA.NextToFill];
  ReadBufB=BufsB.Bufs[BufsB.NextToFill];
  SBufPrepare(ReadBufA);
  SBufPrepare(ReadBufB);
  WriteBuf=NULL;
  PromisedStatus=1; /* send a status block immediately on write pipe */
  

  /* main event loop */
  for (;;) {
    /* set up select() parameters */
    FD_ZERO(&readfd);
    FD_ZERO(&writefd);
    
    maxfd=maxva(InputAPipe,InputBPipe,OutputPipe,GUIRead,GUIWrite->writefd,-2);
    if (ReadBufA && SBufIsReady(ReadBufA))
      FD_SET(InputAPipe,&readfd);
    if (ReadBufB && SBufIsReady(ReadBufB))
      FD_SET(InputBPipe,&readfd);
    if (WriteBuf && SBufIsReady(WriteBuf))
      FD_SET(OutputPipe,&writefd);
    
    if (GUIRead >= 0)
      FD_SET(GUIRead,&readfd);
    if (GUIWrite->writefd >= 0 && queued_outbuffer_data_available(GUIWrite)) {
      FD_SET(GUIWrite->writefd,&writefd);
    }
    
    select(maxfd+1,&readfd,&writefd,NULL,NULL);
    if (FD_ISSET(InputAPipe,&readfd)) {
      /* read from pipe A */
      ReadSBuf(ReadBufA,InputAPipe);
      
    }
    if (FD_ISSET(InputBPipe,&readfd)) {
      /* read from pipe B */
      ReadSBuf(ReadBufB,InputBPipe);
    }
    if (FD_ISSET(OutputPipe,&writefd)) {
      /* write to output pipe */
      WriteSBuf(WriteBuf,OutputPipe);
      
    }
    if (FD_ISSET(GUIRead,&readfd)) {
      sspi_readchars();
    }
    if (FD_ISSET(GUIWrite->writefd,&writefd)) {
      write_queued_outbuffer(GUIWrite);
    }

    /* Now see what's finished */
    
    /* first see if the write buffer is finished */
    if (WriteBuf && SBufIsDone(WriteBuf)) {
      WriteBuf=NULL;
    }
    
    /* Now check the completion of ReadBufA and ReadBufB */
    if (ReadBufA && SBufIsDone(ReadBufA) && SBufIsStatus(ReadBufA)) {
      SBufPrepare(ReadBufA); /* We want to get the next data block too */
      if (!NumAThrowaway) PostStatus(ReadBufA); /* This causes everything to be reset,
			       so we don't need to do anything else 
			       if there is a major parameter change */

      NextAIsStatus=0;
    }
    else if (ReadBufA && SBufIsDone(ReadBufA)) {
      NextAIsStatus=(SBufGetDataHeader(ReadBufA))->NextBlockIsStatusBlk;
      ReadBufA=NULL;
      /* Increment next-to-read pointer. */
      if (!NumAThrowaway) BufsA.NextToFill=MODINC(BufsA.NextToFill,BufsA.NumBufs);
      else NumAThrowaway--;
    }


    if (ReadBufB && SBufIsDone(ReadBufB) && SBufIsStatus(ReadBufB)) {
      SBufPrepare(ReadBufB); /* We want to get the next data block too */
      if (!NumBThrowaway) PostStatus(ReadBufB); /* This causes everything to be reset if there are major changes ,
			       so we don't need to do anything else */      
      NextBIsStatus=0;
    }
    else if (ReadBufB && SBufIsDone(ReadBufB)) {
      NextBIsStatus=(SBufGetDataHeader(ReadBufB))->NextBlockIsStatusBlk;
      ReadBufB=NULL;
      /* Increment next-to-read pointer. */
      if (!NumBThrowaway) BufsB.NextToFill=MODINC(BufsB.NextToFill,BufsB.NumBufs);
      else NumBThrowaway--;

    }

    /* Now see what we should do */
    /* Rules: output NextToOutput if possible.
       Once NextToOutput is more than JumpBackBufs beyond NextToEmpty, we can 
       empty it and increment NextToEmpty       

       If the one beyond NextToOutput is not full in both inputs,
       try to read in both inputs.

       Do not allow the buffers to wrap around (overflow) !!! 
       */

    /* Write if possible (check first if we have to write a status block */
    if (!WriteBuf && PromisedStatus) {
      PromisedStatus=0;
      WriteBuf=WriteStatusBuf;
      SBufSetStatusBlk(WriteBuf,NumChannels,SampleRate,WriteFrameNum,BlockSize);
      SBufPrepareStatus(WriteBuf);
    }
    if (!WriteBuf && InActiveRegion(&BufsA,NextToOutput) && InActiveRegion(&BufsB,NextToOutput)) {
      if (Switch=='A') {
	WriteBuf=BufsA.Bufs[NextToOutput];
      } else {
	WriteBuf=BufsB.Bufs[NextToOutput];
      }

      WriteFrameNum++;
      if (!(WriteFrameNum % StatusBlkSpacing)) {
	PromisedStatus=1;
      } else {
	PromisedStatus=0;
      }
      (SBufGetDataHeader(WriteBuf))->NextBlockIsStatusBlk=PromisedStatus;
      
      SBufPrepare(WriteBuf);
      NextToOutput=MODINC(NextToOutput,BufsA.NumBufs);
      
    }
   
    /* Empty any possible buffers (once we have enough JumpBack data) 
       Must be careful, because we could still be transmitting 
       NextToOutput-1 */
    if (BuffDiff(NextToOutput,BufsA.NextToEmpty,BufsA.NumBufs) > (JumpBackBufs+1)) {
      SBufEmpty(BufsA.Bufs[BufsA.NextToEmpty]);
      BufsA.NextToEmpty=MODINC(BufsA.NextToEmpty,BufsA.NumBufs);
      
      SBufEmpty(BufsB.Bufs[BufsB.NextToEmpty]);
      BufsB.NextToEmpty=MODINC(BufsB.NextToEmpty,BufsB.NumBufs);
      
    }
 
    /* If we need data on either BufsA or BufsB, collect data on both */
    if (!InActiveRegion(&BufsA,NextToOutput) || !InActiveRegion(&BufsB,NextToOutput)) {
      /* If we need to get some data, read in BOTH channels if possible */
      if (!ReadBufA && MODINC(BufsA.NextToFill,BufsA.NumBufs) != BufsA.NextToEmpty) {
	ReadBufA=BufsA.Bufs[BufsA.NextToFill];
	if (NextAIsStatus)
	  SBufPrepareStatus(ReadBufA);
	else SBufPrepare(ReadBufA);
      }
      if (!ReadBufB && MODINC(BufsB.NextToFill,BufsB.NumBufs) != BufsB.NextToEmpty) {
	ReadBufB=BufsB.Bufs[BufsB.NextToFill];
	if (NextBIsStatus)
	  SBufPrepareStatus(ReadBufB);
	else SBufPrepare(ReadBufB);
      }
    }
       
  }

  return 0;
}
